"""Multiprocessing
Module that contains all logic concerning multiprocessing (including GPU access management) and harware info.

Attributes:
    comm_module: Communication module.
    workers: .
    daemons: .
    free_cpus: .
"""
import logging
from itertools import repeat
import sys
import os

#sys.stderr = open(os.devnull, "w")
import multiprocessing
from multiprocessing import cpu_count
from multiprocessing import Process
from multiprocessing import Lock
from multiprocessing import Manager
from multiprocessing import Queue
from multiprocessing import resource_tracker
from multiprocessing.shared_memory import SharedMemory
#sys.stderr = sys.__stderr__
from xframe.Multiprocessing_interfaces import OpenClDependency,PsutilDependency,SharedArrayDependency
from xframe.library.pythonLibrary import xprint
import traceback
import numpy as np
from xframe import settings

sa = SharedArrayDependency #will be dependency injected with SharedArray module
MPI = False #will be dependency injected with SharedArray module
mpi_comm = None #will be set by create_mpi_comm #None if not jet checked False if mpi not available otherwise MPI.COMM_WORLD
comm_module = 'Will be set by Controller.'
openCL_plugin = OpenClDependency
psutil_plugin = PsutilDependency
log = logging.getLogger('root')
master_name = 0

def create_mpi_comm():
    global mpi_comm
    if not isinstance(MPI,bool):
        mpi_comm = MPI.COMM_WORLD # global communicator
    
def get_global_number_of_processes():
    if mpi_comm == None:
        create_mpi_comm()
        
    if isinstance(mpi_comm,bool):        
        global_number_of_processes = 1
    else:
        global_number_of_processes=mpi_comm.Get_size()
        
        #comm=commWorld.Split_type(MPI.COMM_TYPE_SHARED) # Inter node communicator
        #globalRank=commWorld.Get_rank()
        #nodeRank=comm.Get_rank()    
        #nodeNumberOfThreads=comm.Get_size()
        
        #MPIPoolExecutor =  mpi4py.futures.MPIPoolExecutor
        #mpi_executor=MPIPoolExecutor()        
    return global_number_of_processes

def get_local_cpu_count():
    #cpu_count gives number of cpu_threads. To get rid of hyperthreading divide py 2 
    return int(cpu_count()//2)
    
def load_openCL_dict():
    if not isinstance(openCL_plugin,str):
        process_identity = multiprocessing.current_process()._identity
        if len(process_identity)<1:
            process_number = 0
        else:            
            process_number = process_identity[-1]
        #log.info("process number = {}".format(process_number))
        #log.info("process = {}".format(multiprocessing.current_process()))
        #log.info("process identity = {}".format(multiprocessing.current_process()._identity))
        state = openCL_plugin.load_openCL_dict(process_number = process_number)
    else:
        e = AssertionError('openCL plugin has not been injected.')
        raise e
    return state
        
def get_Event():
    return multiprocessing.Event()
def get_Queue(using_manager = False):
    #manager = multiprocessing.Manager()
    if using_manager:
        manager = multiprocessing.Manager()
        return manager.Queue()
    else:
        return multiprocessing.Queue()

def get_process_name():
    current_process = multiprocessing.current_process()
    if len(current_process._identity)<1:
        process_id = master_name
    else:            
        process_id = int(current_process.name)
    return process_id
def get_n_child_processes():
    non_sync_children = tuple(child for child in multiprocessing.active_children() if not 'SyncManager-' in str(child.name))
    return len(non_sync_children)

def get_free_memory():
    return psutil_plugin.get_free_memory()

class MPProcess:
    def __init__(self,process=False,queue=False,events={}):
        if isinstance(process,multiprocessing.Process):
            self.process = process
            self.name = process.name
            self.is_daemon = process.daemon
            self.start = process.start
            self.join = process.join
        else:
            #assume master process is meant
            self.process = 'master'
            self.name = 0
            self.is_daemon = False
            self.start = False
            self.join = False
        self.queue = queue
        self.events = events # dict of Events 

daemon_termination_queue = get_Queue()
master_process = MPProcess(queue = daemon_termination_queue,events={'gpu':get_Event()})

gpu_queues = {i:get_Queue() for i in range(settings.general.max_parallel_processes+1)}
gpu_events = {i:get_Event() for i in range(settings.general.max_parallel_processes+1)}
# Dict of MPProcess objects for worker processes
workers = {}
# Dict of MPProcess objects for daemon processes
daemons = {}

# Active_processes acts as a counter everytime new child processes are started they know how many mp processes where already started when they are called.
free_cpus = get_local_cpu_count()-get_n_child_processes()-1
free_threads = int( 2*get_local_cpu_count()-get_n_child_processes() )-1
default_n_processes = free_cpus

def create_process_names(n,is_daemon=False):
    '''
    Creates process names. There can only be one set of N processes, that are not daemons, active at the same time so give them the names (1,...,N).
    Daemons get unique consequtive negative numbers, the names of killed daemons will not be filled again.
    The name of the master process is by default 0.
    :param n: number of processes to name.
    :param daemon: Flag that specifies whether daemon or worker names should be created.
    :return: tuple of names (positive ingegers for workers, negative integers for daemons)
    '''
    if not is_daemon:
        # If not daemon ruturn (1,...,n)
        names = tuple(range(1,n+1))
    else:
        # If daemon ruturn (lowest_existing_daemon_name-1,...,lowest_existing_daemon-n)
        lowest_existing_name = 0
        for d in daemons.values():
            #log.info('daemon name = {}'.format(d.name))
            if int(d.name)<lowest_existing_name:
                lowest_existing_name = d.name
        names = tuple(range(lowest_existing_name-1,lowest_existing_name-n-1,-1))
    return names

def get_free_cpus():
    return get_local_cpu_count()-get_n_child_processes()-1
def update_free_cpus():
    global free_cpus
    global free_threads
    global default_n_processes    
    free_cpus = get_free_cpus()
    free_threads = int(2*get_local_cpu_count()-get_n_child_processes())-1
    default_n_processes = free_cpus
def check_available_cpus_before_spawning_childs(n_childs,are_daemons=False):
    update_free_cpus()
    would_use_hyperthreads = (free_cpus - n_childs<0)
    would_use_more_processes_than_available_hyper_threads =  (free_threads - n_childs<0)
    if would_use_hyperthreads and (not would_use_more_processes_than_available_hyper_threads):
        if are_daemons:
            log.warning('Total number of daemon processes exceedes CPU core count. Some of the spawned daemons will use hyperthreading threads!')
        else:
            log.warning('Total number of processes exceedes CPU core count. Some of the spawned processes will use hyperthreading threads. This may slow down performance.')

    if would_use_more_processes_than_available_hyper_threads:
        if are_daemons:
            log.warning('Total number of daemon processes exceedes CPU thread count including hyperthreads! This will slow down performance!')
        else:
            log.warning('Total number of processes exceedes CPU thread count including hyperthreads. This will slow down performance.')
    


def _register_processes(processes):
    global daemons
    global workers
    for process in processes:
        if process.is_daemon:
            daemons[process.name]=process
        else:
            workers[process.name]=process
            
def _deregister_processes(processes,exclude_daemons=True):
    global daemons
    global workers
    for process in processes:
        try:
            if process.is_daemon:
                if not exclude_daemons:                
                    daemons.pop(process.name)
            else:
                workers.pop(process.name)
        except KeyError as e :
            log.warning('Process to be deleted with name {} does not exist. Ignoring process deregister event.'.format(name_to_delete))
            
def _process_daemon_terminations():
    while daemon_termination_queue.qsize() != 0:
        name_to_delete = daemon_termination_queue.get()
        try:
            log.info(f"Daemon {name_to_delete} has terminated removing it from daemon list.")
            _deregister_processes([daemons[name_to_delete]],exclude_daemons=False)
        except KeyError:
            log.warning('Daemon to be deleted with name {} does not exist. Ignoring daemon termination event.'.format(name_to_delete))
def get_mp_process_by_name(global_name):
    if global_name<0:
        return daemons[global_name]
    elif global_name>0:
        return workers[global_name]
    else:
        return master_process

def check_if_daemon(global_name:int):
    return global_name<0

def get_process_queue(process_name):
    #log.info('request queue for process {} of {}'.format(process_id,len(queue_per_cpu)))
    process_name = int(process_name)
    try:
        if process_name>0:
            queue = workers[process_name].queue
        elif process_name<0:
            queue = daemons[process_name].queue
        else:
            queue = master_process.queue
    except Exception as e:
        #e = AssertionError('Process {} does not have a queue.'.format(process_name))
        raise e
    return queue
def get_process_event(process_name,event_name=False):
    process_name = int(process_name)    
    try:
        if process_name<0:
            event = daemons[process_name].events[event_name]
        elif process_name>0:
            event = workers[process_name].events[event_name]
        else:
            event = master_process.events[event_name]                            
    except Exception:
        e = AssertionError('process with name {} does not exist or does not have an event called {}'.format(process_name,event_name))
        raise e
    return event
def put_to_process_queue(process_name,message):
    get_process_queue(process_name).put(message)
def get_from_process_queue(process_name,timeout = None):
    return get_process_queue(process_name).get(timeout = timeout)

def get_gpu_queue(process_name):
    #log.info('request queue for process {} of {}'.format(process_id,len(queue_per_cpu)))
    process_name = int(process_name)
    try:
        queue = gpu_queues[process_name]
    except Exception as e:
        raise e
    return queue
def get_gpu_event(process_name):
    process_name = int(process_name)    
    try:
        event = gpu_events[process_name]
    except Exception as e:
        raise e
    return event
    
def put_to_gpu_queue(process_name,message):
    get_gpu_queue(process_name).put(message)
def get_from_gpu_queue(process_name,timeout = None):
    return get_gpu_queue(process_name).get(timeout = timeout)

            

#################################
## Shared Memory (SharedArray) ##

def delete_shared_arrays():
    global sa    
    shared_arrays = sa.list()
    for array in shared_arrays:
        sa.delete(array.name.decode('utf-8'))        

def create_shared_arrays(shapes,dtypes):
    global sa
    names = []
    shared_outputs = []
    #log.info('all output dtypes in Shared array = {}'.format(out_dtypes))
    for i,shape,dtype in zip(np.arange(len(shapes)),shapes,dtypes):
        # The probability of identical hashes from hash(np.random.rand()) should be low enough to not care about it :p 
        name = 'shared_output_'+ str(hash(np.random.rand()))
        names.append(name)
        #log.info('output dtypes in Shared array = {}'.format(out_dtype))
        shared_outputs.append(sa.create('shm://'+name,int(np.prod(shape)),dtype).reshape(shape))
    return shared_outputs, names 


#############################
##  Multiprocessing modes  ##

additional_argument_names = ['number_of_processes','local_name','global_name','events','queue','available_memory','synchronize']


class MPFunctionArgumentsStandard:
    def __init__(self, function, arguments, argument_indices, constant_args = [],call_with_multiple_arguments = False, split_together = False, additional_kwargs = {}):
        self.function = function
        self.arguments = arguments
        self.argument_indices = argument_indices
        self.split_together = split_together
        self.constant_args = constant_args
        self.call_with_multiple_arguments = call_with_multiple_arguments
        self.additional_kwargs = additional_kwargs
        for kw in additional_argument_names:
            assert kw in additional_kwargs, 'expected argument {} not provided.'.format(kw)
        
class MPMode:
    name = 'none'
    def __init__(self):
        self.pre_processed = False
    def pre_processing(self,n_processes):
        self.pre_processed = True
    def get_mp_function(self,mp_arguments:MPFunctionArgumentsStandard):
        assert self.pre_processed , 'call to get_mp_function before mode attributes where processed via pre_process'
    def collect_outputs(self,result,split_dict):
        return result
    def exception_decorator(self,func,additional_kwargs):
        def outer(*args,**kwargs):
            try:
                func(*args,**kwargs)
                # if daemon request deletion from daeomes dict in master process
                global_name = additional_kwargs['global_name']
                if check_if_daemon(global_name):
                    master_process.queue.put(global_name)
                    #log.info('Request deletion of exiting daemon with global_name {}'.format(global_name))                
            except Exception as e:
                traceback.print_exc()
                log.error(e)
                self.exception_handler(e,**additional_kwargs)
            #log.info('calculation done in process {}'.format(additional_kwargs['global_name']))
        return outer
    def exception_handler(self,exception,**kwargs):
        global_name = kwargs['global_name']
        if check_if_daemon(global_name):
            global_name = kwargs['global_name']
            master_process.queue.put(global_name)
            #log.info('Request deletion of failed daemon with global_name {}'.format(global_name))
        else:
            raise exception
                      
class MPMode_Queue(MPMode):
    name = 'Queue'
    def __init__(self,assemble_outputs=True,ignore_first_dimension=False):
        super().__init__()
        self.assemble_outputs = assemble_outputs
        self.queue = get_Queue(using_manager = True)
        self.ignore_first_dimension = ignore_first_dimension
    def get_mp_function(self,mp_arguments:MPFunctionArgumentsStandard):
        assert self.pre_processed , 'call to get_mp_function before mode attributes where processed via pre_process'      
        # handle edge case of a no argument
            
        queue = self.queue
        additional_kwargs = mp_arguments.additional_kwargs
        local_name = additional_kwargs['local_name']
        function = mp_arguments.function
        args = mp_arguments.arguments                
        c_args = mp_arguments.constant_args
        if mp_arguments.call_with_multiple_arguments:        
            def mp_function():
                #log.info(additional_kwargs.keys())
                result=function(*args,*c_args,**additional_kwargs)
                #            log.info('result type={}'.format(type(result)))
                if not isinstance(result,np.ndarray):
                    result=[result]
                    
                result_dict={local_name:result}
                #           log.debug('processID type={}'.format(type(processID)))
                queue.put(result_dict,False)
                
        else:
            def mp_function():
                #log.info('arguments = {}'.format(args))
                #log.info(tuple(zip(*args)))
                if len(args)==0:
                    result=[function(*c_args,**additional_kwargs)]
                else:
                    #log.info(args)
                    n_args=len(args[0])
                    #log.info( list( [[arg[i] for arg in args],c_args,additional_kwargs] for i in range(n_args)))
                    result=list(function(*[arg[i] for arg in args],*c_args,**additional_kwargs) for i in range(n_args))
                #log.info("len result = {}".format(len(result)))
                #combine individual results if possible
                if len(result)>0:
                    all_results_are_numpy_arrays = np.array([isinstance(r,np.ndarray) for r in result]).all()
                    if all_results_are_numpy_arrays:
                        all_results_have_same_shape = np.array([r.shape==result[0].shape for r in result]).all()
                        #log.info('result shapes = {}'.format(np.array([r.shape for r in result])))
                        if all_results_have_same_shape:
                            result = np.stack(result,axis = 0)
                            #log.info('result shape = {}'.format(result.shape))
                #log.info(args)
                #log.info(result[0])
                #if (len(args)<=1) and (len(result[0])<=1):
                #    pass# result = result[0]
                result_dict={local_name:result}
                queue.put(result_dict,False)
                #log.info('put to queue in process {}'.format(additional_kwargs['global_name']))
                
        return self.exception_decorator(mp_function,additional_kwargs)
        
    def collect_outputs(self,split_dict):
        out_dict={}
        queue = self.queue
        while queue.qsize() != 0:
            out_dict.update(queue.get())
        #log.info('assemble outputs = {}'.format(self.assemble_outputs))
        if self.assemble_outputs:
            return self.assemble_results(out_dict,split_dict)
        else:
            #log.info('out_dict keys = {}'.format(out_dict.keys()))
            #log.info('split_dict = {}'.format(split_dict))
            #log.info('shape part = {}'.format(split_dict['shape_part']))
            one_output_per_process = np.prod(split_dict['shape_part'])<=len(split_dict['indices'])
            #log.info('one_ouput_per_process = {}'.format(one_output_per_process))
            #log.info('shape part = {},{}'.format(split_dict['shape_part'],len(split_dict['indices'])))
            if one_output_per_process:
                out_dict = {key:val[0] for key,val in out_dict.items()}
            return out_dict
    def assemble_results(self,result_dict,split_dict):
        '''
        Uses the splitted_indices from _split_arguments and tries to assemble the multiprocessing outputs into an numpy array or list depending on the return type of the first element in the results dict.    
        '''
        #log.info('type of result before assembly = {}'.format(type(result_dict)))
        #log.info('start assembling output')
        splitted_indices = split_dict['indices']
        result_shape_part = split_dict['shape_part']
        split_mode = split_dict['mode']
        #result_shape_end=(max(len(argArrays),1),)
        if len(result_dict)<=0:
            return result_dict
        
        first_result=next(iter(result_dict.values()))
        result_type = type(first_result)        
        
        #log.info('type of result = {}'.format(result_type))
        #log.info(tuple(i.shape for i in next(iter(result_dict.values()))))
        if result_type == np.ndarray:
            if len(first_result.shape)>1 or self.ignore_first_dimension:
                n=1
            else:
                n=0
            first_shape=first_result.shape[n:]
            #log.info('shape part ={} result[0] shape ={}, first shape = {}'.format(result_shape_part,first_result.shape,first_shape))
            shapes_equal = False
            if np.array(tuple(isinstance(r,np.ndarray) for r in result_dict.values())).all():
                shapes_equal = np.array([np.array(r.shape[n:])==np.array(first_shape) for r in result_dict.values()]).all()
            
            if shapes_equal:
                #log.info('shapes are equal !')
                result_shape=result_shape_part+first_shape
                result_dtype=first_result.dtype        
                result=np.empty(result_shape,dtype=result_dtype)
                #log.info(f'result shape = {result.shape} result.dtype = {result.dtype}')
                for process_id in result_dict.keys():
                    if not split_dict['split_together']:
                        result_part_tuple=tuple(indice_part.astype(int) for indice_part in splitted_indices[process_id] )
                    else:
                        result_part_tuple=(splitted_indices[process_id][0],)
                    #log.info('result part tuple = {}'.format(result_part_tuple))
                    result[result_part_tuple]=np.array(result_dict[process_id])
            else:
                #log.info('shapes are NOT equal !')
                result = self._assemble_results_as_lists(result_dict,split_dict)
                
        elif (result_type == tuple) or (result_type==list):
            result = self._assemble_results_as_lists(result_dict,split_dict)
        else:
            result = np.array([])
        #log.info(f'result shape = {result.shape} result.dtype = {result.dtype}')
        return result

    def _assemble_results_as_lists(self,result_dict,split_dict):
        '''
        Assembles results as list
        '''
        #log.info('type of result before assembly = {}'.format(type(result_dict)))
        split_mode = split_dict['mode']
        if len(result_dict)<=0:
            return result_dict
                
        first_result=next(iter(result_dict.values()))
        result_type = type(first_result)
        
        n_processes = len(result_dict)
        if split_mode=='sequential':
            result = []
            for process_id in range(n_processes):
                try:
                    result+=list(result_dict[process_id])
                except KeyError as e:
                    log.error('Process {} seems to have failed. Skipping.'.format(process_id))
        elif split_mode=='modulus':
            parts = len(result_dict[0])
            result_parts = [[] for n in range(parts)]
            
            for process_id in range(n_processes):
                try:
                    for part,output in enumerate(result_dict[process_id]):
                        result_parts[part].append(output)
                except KeyError as e:
                    log.error('Process {} seems to have failed. Skipping.'.format(process_id))
            result = [item for sub_list in result_parts for item in sub_list]
        else:
            raise AssertionError('split_mode = {} not known.'.format(split_mode))
        return result
    
class MPMode_SharedArray(MPMode):
    name = 'SharedArray'
    def __init__(self,output_shapes,output_dtypes,reduce_arguments=False):
        super().__init__()
        self._initial_output_shapes = output_shapes
        self.output_shapes = output_shapes
        self.output_dtypes = output_dtypes
        self.reduce_arguments = reduce_arguments
        
        ## The following attributes will be set by pre_processing
        self.shared_outputs = False
        self.output_names = False
    def pre_processing(self,n_processes):
        ## delete any existing and create new shared memory objects
        delete_shared_arrays()
        if self.reduce_arguments:
            new_shapes = [(n_processes,)+shape for shape in self._initial_output_shapes]
            self.output_shapes = new_shapes
                
        self.shared_outputs,self.output_names = create_shared_arrays(self.output_shapes,self.output_dtypes)
        self.pre_processed = True
    def _generate_output_indices(self,mp_arguments):
        local_name = mp_arguments.additional_kwargs['local_name']

        output_indices=tuple(indice_part.astype(int) for indice_part in mp_arguments.argument_indices)
            
        if mp_arguments.split_together:
            output_indices = output_indices[0]
        if self.reduce_arguments:
            output_indices = slice(local_name,local_name+1)
        return output_indices
        
    def get_mp_function(self,mp_arguments:MPFunctionArgumentsStandard):
        assert self.pre_processed , 'call to get_mp_function before mode attributes where processed via pre_process'
        #repeat constant arguments
        repeated_arguments=[repeat(arg) for arg in mp_arguments.constant_args]

        # handle edge case of a no argument
        args = mp_arguments.arguments
        if len(args)==0:
            args=[[None]]
            
        additional_kwargs = mp_arguments.additional_kwargs
        local_name = additional_kwargs['local_name']
        output_names = self.output_names
        output_shapes = self.output_shapes
        output_indices = self._generate_output_indices(mp_arguments)

        args = mp_arguments.arguments
        c_args = mp_arguments.constant_args
        function = mp_arguments.function
        
        if mp_arguments.call_with_multiple_arguments:        
            def mp_function():            
                output_files = [sa.attach('shm://'+ o_name).reshape(o_shape) for o_name,o_shape in zip(output_names,output_shapes)]
                additional_kwargs['outputs'] = output_files
                additional_kwargs['output_ids'] = output_indices
                # in this function output have to be placed in outputs[i][output_ids]
                function(*args,*c_args,**additional_kwargs)            
        else:
            def mp_function():
                output_files = [sa.attach('shm://'+o_name).reshape(o_shape) for o_name,o_shape in zip(output_names,output_shapes)]
                additional_kwargs['outputs'] = output_files
                for argument,out_ids in zip(zip(*args),zip(*output_indices)):
                    additional_kwargs['output_ids'] = out_ids
                    #In this function output have to be placed in outputs[i][output_ids]
                    function(*argument,*c_args,**additional_kwargs)
        return self.exception_decorator(mp_function,additional_kwargs)            

    def collect_outputs(self,split_dict):
        # The following only affects the ability to share the arrays it does not cause them to be physically deleted untill the last reference vanishes.
        delete_shared_arrays()
        return self.shared_outputs
    
    
def _run_processes(process_list):
    for p in process_list:
        p.start()
        
def _join_processes(process_list):
    for p in process_list:
        if not p.is_daemon:
            p.join()
            #log.info('process {} joined'.format(p.name))

def split_mp_arguments(argArrays,nProcesses,mode='sequential',split_together=False):
    """
    Routine that splits the list of argArrays into as nProcesses maky pieces.
    mode can either be 'sequential' or 'modulus'. In sequential mode an input array is split into sequential pieces,
    e.g. argArrays =[[1,2,3,4,5]] would be split for nProcesses=2 into [1,2],[3,4,5] (note that the last pice can be bigger by nProcesses-1 elements than the other parts).
    In 'modulus' the arguments are split in to equivalence classes of $\mathbb{Z}/nProcesses \mathbb{Z}$,
    e.g: argArrays =[[1,2,3,4,5]] would be split for nProcesses=2 into [1,3,5],[2,4] (in this case all part sizes can differ at most by 1).
    
    If more than 1 array is given as input by default all possible parameter combinations are considered and splitted into pices,
    e.g.: argArrays =[[1,2],[2,3]] is considered as [(1,2),(1,3),(2,2),(2,3)] which is then split into parts.
    If however split_together = True all input arrayys have to be of the same length and are split together,
    e.g.: argArrays =[[1,2],[2,3]] is considered as [(1,2),(2,3)] and then split into parts.
    
    This method not only returns the splitted arguments but also their corresbonding indices as part of an (len_argument_array_1,...,len_argument_array_N) shaped  array,
    which in case of splitTogether= true reduces to (len_argument_array_1,) rhaped array.
    """
    
    #log.info('argArrays={}'.format(argArrays))
    if isinstance(argArrays,np.ndarray):
    #    log.info('yep')
        argArrays=[argArrays]
    indiceArrays=[np.arange(len(array),dtype=int) for array in argArrays]
    arg_legths_equal= len(set(tuple(len(array) for array in argArrays))) == 1
    dtype_list = [arg.dtype for arg in argArrays]
    n_args = len(argArrays)
    #log.info('len argArrays ={}'.format(len(argArrays)))
    if split_together:
        try:
            assert arg_legths_equal,'split_together requires all argArrays to have the same length but the length are {}'.format(tuple(len(array) for array in argArrays)) 
        except AssertionError as e:
            log.error(e)
            raise
        arguments=tuple(argArrays)
        argIndices=indiceArrays
       
    else:
        #    log.info('indiceArrays={}'.format(indiceArrays))
        argIndices=tuple(reversed(tuple(dataset.flatten().astype(np.int_) for dataset in np.meshgrid(*reversed(indiceArrays)) )))
        arguments=tuple(argument[indices] for argument,indices in zip(argArrays,argIndices))
    nArguments=len(arguments[0])
    #print('arguments = {}'.format(argArrays))    
    #log.info('nArgs = {} nProcess {}'.format(nArguments,nProcesses))
    if nArguments<nProcesses:
        nProcesses=nArguments
    argPerProcess=nArguments//nProcesses    

    #    log.info('nArguments={}, nProcesses={}'.format(nArguments,nProcesses))
    #print(mode)

    if mode=='sequential':    
        splitPoints=[int(argPerProcess*process) for process in range(1,nProcesses)]
        #splittedArguments=tuple(np.split(argument,splitPoints) for argument in arguments)
        splittedArguments=tuple(np.array_split(argument,nProcesses) for argument in arguments)
        #splittedIndices=tuple(np.split(argIndice,splitPoints) for argIndice in argIndices)
        splittedIndices=tuple(np.array_split(argIndice,nProcesses) for argIndice in argIndices)
    elif mode=='modulus':
        def splitArgument(argument):
            dtype=argument.dtype
            def returnSlizedArgument(processID):
                arg_slice = np.array(argument[processID::nProcesses],dtype=dtype)
                #log.info('slice shape = {}'.format(arg_slice.shape))
                return arg_slice
            splitted_argument = tuple(map(returnSlizedArgument,range(nProcesses)))
            return splitted_argument
        
        splittedArguments=tuple(splitArgument(argument) for argument in arguments)
        splittedIndices=tuple(splitArgument(argIndice) for argIndice in argIndices)
        #log.info('dtype={} splitted Indices=\n {}'.format(splittedIndices.dtype,splittedIndices))
    else:
        raise AssertionError('Split mode "{}" not known. Known modes are "sequential" or "modulus".'.format(mode))
    arguments_per_process = list(zip(*splittedArguments))
    indices_per_process = list(zip(*splittedIndices))    
    return {'arguments':arguments_per_process,'indices':indices_per_process}

def _create_synchronization_routines(process_ids):
    '''
    Generates a synchronization routine to be used in shared memory multiprocessing calls.
    It allows for all participating processes to synchronize i.e. wait untill all have finished a given task.
    Implementateion details:
    Each process has a pair of events which function as a toggle between states (0,1) and (1,0)
    A synchronization is equivalent to wait until all processes have switched their state once.
    
    If a process fails to synchronice once (time out is reached), it will be excluded from all further synchronization attempts, which means:
    1. If it tries to synchronice an AssertionError is thrown.
    2. All other Processes ignore its state in further synchronization attempts.
    :param process_ids: names of the processes for which to create synchronization functions
    :param type: list of different hashable elements
    :return: Dictionary of synchronization functions whose keys are the corresponding process_ids. A call to any function blocks untill timout is reached or all other functions are called.
    :return type: dict
    '''
    #create events
    process_events = {p:[get_Event(),get_Event()] for p in process_ids}
    # initialize events to be in state (0,1)
    for events in process_events.values():
        events[1].set()

    def process_id_decorator(local_process_id):        
        def inner(func):
            def wrapper(timeout=30):
                return func(local_process_id,timeout=timeout)
            return wrapper
        return inner

    routines = {}
    for p in process_ids:
        @process_id_decorator(p)
        def synchronize(local_process_id,timeout=30):
            #log.info('timeout = {}'.format(timeout))
            # get local id of zero state event (is either 0 or 1)
            zero_id = int(process_events[local_process_id][0].is_set())
            one_id = (zero_id+1)%2
            local_event = process_events[local_process_id]
            has_failed = local_event[zero_id].is_set() == local_event[one_id].is_set()

            # Check wether local process did fail to synchronice in the past (i.e is in state 00 or 11)
            if has_failed:
                error = AssertionError('Synchronization Error: This process {} has failed to synchronice in the past and is thus prevented from further execution.')
                raise(error)
            else:            
                # Set flag that local process awaits toggle switch.
                process_events[local_process_id][zero_id].set()
            
            
            # wait until all other workers signal that they await toggle switch
            for p_id,event in process_events.items():
                event_in_failed_state = ( event[0].is_set()==event[1].is_set() )
                if (p_id == local_process_id) or event_in_failed_state:
                    # don't wait for myself or previousely failed processes 
                    continue
                success = event[zero_id].wait(timeout = timeout)
                if not success:
                    log.warning('Process {} didnt synchronice after {} seconds stop waiting in process {}'.format(p_id,timeout,local_process_id))
                    # set failed process in failed mode:
                    event[zero_id].clear()
                    event[one_id].clear()
                    
            # finalize local toggle switch
            process_events[local_process_id][one_id].clear()
            #log.info('sync_done process {}'.format(local_process_id))
        routines[p]=synchronize
    return routines
                
def _read_number_of_processes(n_processes):
    '''
    If number of processes is of the wrong type this method sets it to the default number.
    If number of processes is bigger than the default default_n_processes it is set to the default.
    The default is get_local_cpu_count()-active_processes.
    '''
    #log.info(f' input n processes = {n_processes}')
    update_free_cpus()
    global free_cpus    
    if isinstance(n_processes,bool) or (not isinstance(n_processes,int)):
        n_processes = free_cpus
    else:
        n_processes = n_processes
    #if n_processes > free_cpus:
    #    n_processes = free_cpus
    n_processes = max(1,n_processes)
    return n_processes

def _split_arguments(arg_arrays,n_processes,mode,split_together):
    '''
    Wrapper around split_mp_arguments, that deals with the case of no argument_arrays beeing provided.
    '''        
    if len(arg_arrays)!=0:
        split_dict = split_mp_arguments(arg_arrays,n_processes,mode=mode,split_together=split_together)
        splitted_args = split_dict['arguments']
        splitted_indices= split_dict['indices']
        #            log.info('splittedIndices 1=/n{}'.format(tuple(splittedIndices)))
        #           log.info('splitted Args shape={}'.format(splittedArgs.shape))
        n_processes=len(splitted_args)
        if split_together:
            result_shape_part=(len(arg_arrays[0]),)
        else:
            result_shape_part=tuple([len(array) for array in arg_arrays])
    else:
        splitted_args=[[]]*n_processes
        splitted_indices=np.expand_dims(np.arange(n_processes),1)
        result_shape_part=(n_processes,)
    return {'arguments':splitted_args,'indices':splitted_indices,'shape_part':result_shape_part,'mode':mode,'split_together':split_together}

def _calc_available_mem_per_process(n_new_processes):
    assert get_process_name() == 0, 'available mem can only be calculated from master process.'
    # create multiprocessing wrapers 
    free_mem=get_free_memory()# virtual_memory()
    available_memory_per_process=free_mem/(n_new_processes + get_n_child_processes())
    return available_memory_per_process

def process_mp_request(function,
                       mode:MPMode = MPMode_Queue(),
                       is_daemon=False,
                       return_process_objects = False,
                       input_arrays=[],
                       const_inputs=[],
                       call_with_multiple_arguments=False,
                       split_mode='sequential',
                       split_together=False,
                       n_processes = False,
                       event_names=[],
                       create_queues = False):
    '''
    '''
    process_name = get_process_name()
    #log.info("mode = {}".format(function))
    assert process_name == master_name , 'Multi processing requests can only be sent from the master process whose name is {}. This request has been sent by process {}. Abborting MP evaluation.'.format(master_name,process_name)
    assert isinstance(mode,MPMode), 'Multiprocessing mode {} is not known. Available modes are {}'.format(mode,tuple(MPMode_Queue,MPMode_SharedArray))

    ## look for daemons that terminated
    _process_daemon_terminations()
    
    ## split input arguments 
    n_processes = _read_number_of_processes(n_processes)
    #log.info('split arguments')
    split_dict = _split_arguments(input_arrays,n_processes,split_mode,split_together)
    n_processes = len(split_dict['arguments'])    
    log.info(f'Starting {n_processes} process{"es"*(n_processes>1)}.')
    log.debug(f'On function: {function}')

    ## Throw warnings if active processes will use HyperThreads
    check_available_cpus_before_spawning_childs(n_processes,are_daemons=is_daemon)


    mode.pre_processing(n_processes)
    
    ## Build Processes
    process_names = create_process_names(n_processes,is_daemon=is_daemon)
    #log.info('New process names = {}'.format(process_names))
    sync_functions = _create_synchronization_routines(process_names)
    available_mem = _calc_available_mem_per_process(n_processes)
    mp_processes = []
 
    for local_name,global_name in enumerate(process_names):
        if create_queues:
            #log.info('creating queue for process {}'.format(global_name))
            queue = get_Queue()
        else:
            queue = False        
        events = {e_name:get_Event() for e_name in event_names}
        additional_kwargs = {}
        additional_kwargs['number_of_processes'] = n_processes
        additional_kwargs['local_name'] = local_name
        additional_kwargs['global_name'] = global_name
        additional_kwargs['events'] = events
        additional_kwargs['queue'] = queue
        additional_kwargs['available_memory'] = available_mem
        additional_kwargs['synchronize'] = sync_functions[global_name]

        arguments = split_dict['arguments'][local_name]
        argument_indices = split_dict['indices'][local_name]
        mp_args = MPFunctionArgumentsStandard(function, arguments, argument_indices, constant_args = const_inputs, call_with_multiple_arguments=call_with_multiple_arguments, split_together=split_together, additional_kwargs=additional_kwargs)
        mp_function  = mode.get_mp_function(mp_args)
        
        if is_daemon:
            log.info(f"Started global daemon with name {global_name}")
            process=Process(target = mp_function, name = global_name, daemon = True)
        else:
            process=Process(target = mp_function, name = global_name)
        mp_process=MPProcess(process, queue = queue, events = events)
        
        mp_processes.append(mp_process)    
    ## update_process_dicts
    _register_processes(mp_processes)
    
    ## Run and join Processes
    _run_processes(mp_processes)
    _join_processes(mp_processes)
    
    outputs = mode.collect_outputs(split_dict)

    _deregister_processes(mp_processes)
    if not is_daemon:
        #log.info('All processes done!'.format(n_processes))
        pass
    if return_process_objects:
        return outputs,mp_processes
    else:
        return outputs
    

#################
##  GPU Access ##
def get_number_of_gpus():
    ###### IMPORTANT ######
    # Due to pyOpenCL not beeing fork save one needs to querry the number of available GPUs in a child process. To avoid future opencl imports in child processes to fail.
    def callback(**kwargs):
        return openCL_plugin.get_number_of_gpus()
    n_gpus = process_mp_request(callback,n_processes =1)[0]
    return n_gpus
def create_openCL_context():
    try:
        openCL_plugin.create_context()
    except Exception as e:
        log.info(f'Could not create openCL context in process {get_process_name()}')
        raise e



def remove_shm_from_resource_tracker():
    """Monkey-patch multiprocessing.resource_tracker so SharedMemory won't be tracked

    More details at: https://bugs.python.org/issue38119
    """

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(self, name, rtype)
    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(self, name, rtype)
    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]
        

class SharedMemoryObject:
    def __init__(self,name,shape,dtype):
        self.name = name
        self.shape = shape
        self.dtype = dtype
        self.nbytes = np.dtype(self.dtype).itemsize * np.prod(self.shape)
        self.shm = False
        self.array = False
        self.created = False
    def release(self):
        #log.info('Trying to delete shared memory object {} P {}'.format(self.name,get_process_name()))
        if isinstance(self.shm,SharedMemory):
#            log.info('trying to delete shared memory object {} P {}'.format(self.name,get_process_name()))
            try:
                self.shm.close()
                #xprint('Shared memory object {} is beeing closed P {}'.format(self.name,get_process_name()))
                if hasattr(self, 'array'):
                    del(self.array)
                if self.created:
                    ## IMPORTANT COMMENT ##
                    #xprint('Shared memory object {} is beeing unlinked P {}'.format(self.name,get_process_name()))
                    self.shm.unlink() #causes a lot of warnings see bugs.python.org/issue38119
                    # unlink is implemented as: 
                    #def unlink(self):
                    #    """Requests that the underlying shared memory block be destroyed.
                    #    In order to ensure proper cleanup of resources, unlink should be
                    #    called once (and only once) across all processes which have access
                    #    to the shared memory block."""
                    #    if _USE_POSIX and self._name:
                    #        _posixshmem.shm_unlink(self._name)
                    #        resource_tracker.unregister(self._name, "shared_memory")
                    # See https://github.com/python/cpython/blob/3.11/Lib/multiprocessing/shared_memory.py
                    # calling both shm_unlink and unregister or just shm_unlink causes a lot of resouce_tracker warnings so untill the bug is fixed
                    # we can only manually unregister the shared memory without calling shm_unlink that gets rid of most of the warnings.
                    # I somehow was not able to suppress the warnings using warnings.simplefilter only by using the python interpreter option -W
                    # which unfortunatelly is not an option in setuptools entry_points.
                    # Well maybe there is an other way that I didn't find.
                    #resource_tracker.unregister(self.shm._name,'shared_memory')                    
                    self.created = False
                    self.shm = False
            except FileNotFoundError as e:
                log.error(e)
                #pass
                #traceback.print_exc()
                
    def construct(self):
        #log.info('trying to create memory object shared {} in P {}'.format(self.name,get_process_name()))
        if not isinstance(self.shm,SharedMemory):
            try:
                self.shm=SharedMemory(self.name,create=True,size=self.nbytes)
            except FileExistsError as e :
                log.warning('left over shared memory file {}. Trying to overwrite. on process {}'.format(self.name,get_process_name()))
                shm=SharedMemory(self.name)
                shm.close()
                shm.unlink()
                self.shm=SharedMemory(self.name,create=True,size=self.nbytes)
                #log.info('create shared memory object {} of size {}'.format(self.name,self.nbytes))
            #resource_tracker.unregister(self.shm._name,'shared_memory') ## prevents some Warnings due to bugs.python.org/issue38119. However it causes a FileNotFoundError when trying to unlik the shared array ...
            self.created = True
            #log.info('create shared memory object {} of size {}'.format(self.name,self.nbytes))
        self.array = np.ndarray(self.shape, dtype=self.dtype, buffer=self.shm.buf)
        return self
    def bind(self,unregister=False):
        #log.info('trying to bind to shared memory object {}'.format(self.name))
        if not isinstance(self.shm,SharedMemory):
            if unregister:
                remove_shm_from_resource_tracker()                
            self.shm=SharedMemory(self.name)
        self.array = np.ndarray(self.shape, dtype=self.dtype, buffer=self.shm.buf)
        return self

class GpuProcessHandle:
    def __init__(self,gpu_process,python_process_id):
        self.gpu_process = gpu_process
        self.python_process_id = python_process_id
        self.inputs,self.outputs = self.create_IO_instances()        
    def create_IO_instances(self):
        gpu_process = self.gpu_process
        inputs = []
        outputs = []
        for i in range(gpu_process.n_inputs):
            in_name = self.generate_shared_memeory_input_name(i,self.python_process_id)
            in_shape = gpu_process.input_shapes[i]
            in_dtype = gpu_process.input_dtypes[i]
            inputs.append(SharedMemoryObject(in_name,in_shape,in_dtype))
        for i in range(gpu_process.n_outputs):
            out_name = self.generate_shared_memeory_output_name(i,self.python_process_id)
            out_shape = gpu_process.output_shapes[i]
            out_dtype = gpu_process.output_dtypes[i]
            outputs.append(SharedMemoryObject(out_name,out_shape,out_dtype))
        return tuple(inputs),tuple(outputs)
    def IO_construct_shared_memory(self):
        for output in self.outputs:
            output.construct()
            #log.info(output.shm)
        for _input in self.inputs:
            _input.construct()
    def generate_shared_memeory_input_name(self,input_id,python_process_id):
        #return self.gpu_process.name+'-hash_{}'.format(self.gpu_process.hash)+'_input_'+str(input_id)+'_process_'+str(python_process_id)
        return self.gpu_process.name+'_input_'+str(input_id)+'_process_'+str(python_process_id)
    def generate_shared_memeory_output_name(self,input_id,python_process_id):
        #return self.gpu_process.name+'-hash_{}'.format(self.gpu_process.hash)+'_output_'+str(input_id)+'_process_'+str(python_process_id)
        return self.gpu_process.name+'_output_'+str(input_id)+'_process_'+str(python_process_id)                                             
    def create_client_function(self,put_to_queue):
        event = get_gpu_event(self.python_process_id)        
        wait_for_done_event=event.wait
        reset_event_flag=event.clear
        input_arrays = tuple(_input.bind(unregister=True).array for _input in self.inputs)
        output_arrays = tuple(_output.bind(unregister=True).array for _output in self.outputs)
        copyto = np.copyto
        copy = np.array
        input_range = range(len(input_arrays))
        output_range = range(len(output_arrays))
        _hash = self.gpu_process.hash

        request_arg=['eval_gpu_process',self.python_process_id,_hash]
        #log.info('input arrays = {}'.format(len(input_arrays)))
        #log.info('output arrays = {}'.format(len(output_arrays)))
        if len(output_arrays)==1:
            output_array=output_arrays[0]
            if len(input_arrays)==1:
                input_array= input_arrays[0]
                def f(arg):
                    #log.info('yay')
                    #log.info('trying to copy to shared memory')
                    #log.info('input array shape = {} shared memory shape = {}'.format(arg.shape,input_array.shape))
                    #log.info('input array shape = {} shared memory shape = {}'.format(arg.dtype,input_array.dtype))
                    #log.info('input array = {}'.format(input_array[:]))
                    #input_array[:] =
                    #log.info(arg)                    
                    copyto(input_array,arg)
                    #log.info('trying to send eval request to controller')
                    #log.info('requesting calculation from Process {}'.format(self.python_process_id))
                    put_to_queue(request_arg)
                    #log.info('Waiting for controller to finish')
                    wait_for_done_event()
                    #log.info('Client {} got done event.'.format(self.python_process_id))
                    #log.info('Unsetting event flag.')
                    reset_event_flag()
                    #log.info('Client {} reset event flag.'.format(self.python_process_id))
                    
                    #log.info('copy output array and return')
                    return copy(output_array)
            else:
                def f(*args):
                    for i in input_range:                        
                        copyto(input_arrays[i],args[i])
                    put_to_queue(request_arg)
                    wait_for_done_event()
                    reset_event_flag()
                    return copy(output_array)                
        else:
            if len(input_arrays)==1:
                input_array= input_arrays[0]
                def f(arg):
                    copyto(input_array,arg)
                    put_to_queue(request_arg)
                    wait_for_done_event()
                    reset_event_flag()
                    return tuple(copy(output_arrays[o]) for o in output_range)
            else:
                def f(*args):
                    for i in input_range:                        
                        copyto(input_arrays[i],args[i])
                    put_to_queue(request_arg)
                    wait_for_done_event()
                    reset_event_flag()
                    return tuple(copy(output_arrays[o]) for o in output_range)
        self.run = f
        return self

    def create_server_function(self):
        self.IO_construct_shared_memory()
        #log.info(self.inputs)
        #log.info([inp.shm for inp in self.inputs])
        #log.info('create_server_function {}'.format(self.python_process_id))
        event = get_gpu_event(self.python_process_id)
        set_event_flag=event.set
        input_arrays = [inp.array for inp in self.inputs]
        output_arrays = [outp.array for outp in self.outputs]
        run_process = self.gpu_process.assemble(input_arrays = input_arrays,output_arrays = output_arrays).run
        def fct():
            #log.info('run server for process {}'.format(self.python_process_id))
            run_process()
            #log.info('process {} server: done!'.format(self.python_process_id))
            set_event_flag()
        self.run = fct
        return self
    
    def remove_shared_memory(self):
        for i in self.inputs:
            #log.info('call __del__ of input {}'.format(i))
            i.release()
        for o in self.outputs:
            #log.info('call __del__ of output {}'.format(o))
            o.release()



class GpuProcessManager:
    @staticmethod
    def request_gpu_handle_client_old(cls,target_name,gpu_process):
        target_queue = get_process_queue(target_name)
        python_process_id = get_process_name()
        client_queue = get_gpu_queue(python_process_id)
        log.info('Sending question of gpu process exists from process {}'.format(python_process_id))
        target_queue.put(['does_gpu_process_exist',python_process_id,gpu_process.hash])
        gpu_process_exists = client_queue.get(timeout=60.0) #tries for 60 seconds 
        log.info('Got answer: {}'.format(gpu_process_exists))
        if not gpu_process_exists:
            log.info('Sending gpu_porocess creation request from id {}'.format(python_process_id))
            target_queue.put(['create_gpu_process',python_process_id,gpu_process.dict])
            gpu_creation_done = client_queue.get(timeout=60.0) #tries for 60 seconds 
            log.info('Gpu process creation is done = {}'.format(gpu_creation_done))
        process_handle = GpuProcessHandle(gpu_process,python_process_id)
        process_handle.create_client_function(target_queue.put)
        log.info('Return client handle for gpu process')
        return process_handle
        
    def __init__(self):
        self.process_dict={}
        self.server_functions={}
        self.process_handles={}
        eval_gpu_process = self.generate_eval_gpu_function_request()
        self.request_dict={
            'does_gpu_process_exist' : self.process_does_gpu_process_exist_request,
            'create_gpu_process':self.process_create_gpu_function_request,
            'eval_gpu_process':eval_gpu_process
        }        
        self.process_request = self.generate_process_request()
        self.client_handles = {} # collects handles in individual client processes        
        
    def request_gpu_handle_client(self,target_name,gpu_process):

        # Check if client handle exists
        _hash = gpu_process.hash
        client_handle_exists=False
        client_handle = self.client_handles.get(_hash,False)
        client_handle_exists = False
        if isinstance(client_handle,GpuProcessHandle):
            if client_handle.python_process_id == get_process_name():
                client_handle_exists = True
        #xprint(f'P{get_process_name()}: Client handle exists: {client_handle_exists}')

        # Check if Server handle exists
        target_queue = get_process_queue(target_name)
        python_process_id = get_process_name()
        client_queue = get_gpu_queue(python_process_id)
        #xprint(f'Sending question if gpu process exists from process {python_process_id} with hash {gpu_process.hash}')
        target_queue.put(['does_gpu_process_exist',python_process_id,gpu_process.hash])
        server_gpu_process_exists = client_queue.get(timeout=60.0) #tries for 60 seconds 
        #log.info('Got answer: {}'.format(server_gpu_process_exists))
        
        if not server_gpu_process_exists:
            #log.info('Sending gpu_porocess creation request from id {}'.format(python_process_id))
            target_queue.put(['create_gpu_process',python_process_id,{"process_dict":gpu_process.dict,"hash":gpu_process.hash}])
            gpu_creation_done = client_queue.get(timeout=60.0) #tries for 60 seconds 
            #log.info('Gpu process creation is done = {}'.format(gpu_creation_done))
        
            client_handle = GpuProcessHandle(gpu_process,python_process_id)
            client_handle.create_client_function(target_queue.put)
            self.register_client_function(client_handle)
        elif not client_handle_exists:
            #xprint("generating client handle")
            client_handle = GpuProcessHandle(gpu_process,python_process_id)
            client_handle.create_client_function(target_queue.put)
            self.register_client_function(client_handle)
        
        #log.info('Return client handle for gpu process')
        #xprint(f"P{get_process_name()}: handle inputs = {client_handle}")
        #xprint(f"P{get_process_name()}: handle inputs = {client_handle.run}")
        return client_handle
        
    def generate_eval_gpu_function_request(self):
        functions_dict = self.server_functions
        def eval_gpu_function(python_process_id,_hash):
            #log.info('eval gpu_func {} for process {}'.format(_hash,python_process_id))
            functions_dict[_hash][python_process_id]()
        return eval_gpu_function
            
    def generate_process_request(self):
        get_request_function = self.request_dict.get
        def do_nothing(python_process_id,data):
            pass
        def process_request(request_type,python_process_id,data):
            get_request_function(request_type,do_nothing)(python_process_id,data)
        return process_request
    def process_create_gpu_function_request(self,python_process_id,request_data):
        gpu_process_data = request_data["process_dict"]
        _hash = request_data["hash"]
        # check again if GPU process was not already created
        server_process_missing = _hash not in self.process_dict
        #xprint(f"Server add gpu: process is still missing {server_process_missing}")
        if server_process_missing:        
            # python_process_id is name of client process wanting to get access to a gpu process.
            gpu_process=openCL_plugin.ClProcess(gpu_process_data)
            self.add_gpu_process(gpu_process)            
            self.register_server_function(gpu_process.hash,python_process_id)
        elif python_process_id not in self.process_handles[_hash]:
            self.register_server_function(_hash,python_process_id)

        #send process created to client
        put_to_gpu_queue(python_process_id,True)
        
    def process_does_gpu_process_exist_request(self,python_process_id,_hash):
        gpu_process_exists = False
        #xprint(f"server hash {_hash} known = {_hash in self.process_dict}")
        if _hash in self.process_dict:
            server_function_for_python_process_is_already_registered =  python_process_id in self.process_handles[_hash]
            #xprint(f"server function exsists = {server_function_for_python_process_is_already_registered}")
            if not server_function_for_python_process_is_already_registered:
                self.register_server_function(_hash,python_process_id)
            
            gpu_process_exists = True
        put_to_gpu_queue(python_process_id,gpu_process_exists)
        #log.info('Process {}, gpu_function already exists {}.'.format(python_process_id,gpu_process_exists))
                        
    def add_gpu_process(self,gpu_process):
        _hash = gpu_process.hash
        process_dict = self.process_dict
        if _hash not in process_dict:
            process_list = openCL_plugin.create_process_buffers_on_all_gpus(gpu_process.dict)
            self.process_dict[_hash]=process_list
    def register_client_function(self,gpu_handle):
        #log.info('added client handle for gpu_process {} on cpu_process{}.'.format(gpu_handle.gpu_process.name,gpu_handle.python_process_id))
        self.client_handles[gpu_handle.gpu_process.hash] = gpu_handle
        
    def register_server_function(self,_hash,python_process_id):
        gpu_process = self.select_gpu_process(_hash,python_process_id)
        handle = GpuProcessHandle(gpu_process,python_process_id)
        handle = handle.create_server_function()
        server_function = handle.run
        functions_by_gpu_process = self.server_functions.get(_hash,{})
        handles_by_gpu_process = self.process_handles.get(_hash,{})
        functions_by_gpu_process[python_process_id]=server_function
        handles_by_gpu_process[python_process_id]=handle
        #xprint(f'registering server function with hash {_hash}' )
        self.server_functions[_hash]=functions_by_gpu_process
        self.process_handles[_hash]=handles_by_gpu_process
        #xprint(f'Server {get_process_name()}: added handle for gpu_process {handle.gpu_process.name} for client  {python_process_id}.')
        
    def select_gpu_process(self,process_hash,python_process_id):
        gpu_process_id = self.distribution_function(process_hash,python_process_id)
        return self.process_dict[process_hash][gpu_process_id]
        
    def distribution_function(self,process_hash,process_id):
        n_func_instances = len(self.process_dict[process_hash])
        return (process_id//settings.general.n_control_workers) % n_func_instances


    def release_shared_memory(self):
        #log.info('process handles {}, client_handles {}'.format(self.process_handles,self.client_handles))
        #log.info('trying to delete gpu process manager in P {}'.format(get_process_name()))
        for h_processes in self.process_handles.values():
            for h in h_processes.values():
                h.remove_shared_memory()
        for h in self.client_handles.values():
            h.remove_shared_memory()
    def remove_shared_memory(self):
        self.__del__()
        





import abc

class SharedMemoryInterface(abc.ABC):

    #_dict structure is as follows
    # {mem_name:{'shape': <tuple(dim1,...)>,'dtype': <np.dtype object >}}
    @abc.abstractmethod
    def allocate(self,_dict):
        pass
    @abc.abstractmethod
    def delete(self,mem_names):
        pass
    @abc.abstractmethod
    def attach(self,mem_names):
        pass
    @abc.abstractmethod
    def detach(self,mem_names):
        pass    
    @abc.abstractmethod
    def __getitem__(self,index):
        pass
    @abc.abstractmethod
    def __len__(self):
        pass
    @abc.abstractmethod
    def __iter__(self):
        pass
    
        



import numpy as np
import numpy.ma as mp
import scipy.integrate as spIntegrate
import logging
from itertools import repeat

log=logging.getLogger('root')

from .hankel_transforms import generate_ht
from .misk import _get_reciprocity_coefficient
from .hankel_transforms import generate_weightDict
from xframe.library.mathLibrary import SphericalIntegrator,PolarIntegrator

#from xframe import database


def load_fourier_transform_weights(dimensions,ft_opt,grid_opt,database):
    db = database
    ft_type = ft_opt['type']
    max_order=grid_opt.max_order
    n_radial_points=grid_opt.n_radial_points
    n_orders=max_order+1
    reciprocity_coefficient = _get_reciprocity_coefficient(ft_opt)
    
    name_postfix='N'+str(n_radial_points)+'mO'+str(max_order)+'nO'+str(n_orders)+'rc'+str(reciprocity_coefficient)
    log.info(f'ft name postfix = {name_postfix}')
    try:
        weights_dict = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimensions)+'D'})
    except FileNotFoundError as e:
        if ft_opt.allow_weight_calculation:
            weights_dict = generate_weightDict(max_order, n_radial_points,reciprocity_coefficient=reciprocity_coefficient,dimensions=dimensions,mode=ft_type)
            if ft_opt.allow_weight_saving:
                db.save('ft_weights',weights_dict,path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimensions)+'D'})
                    
    return weights_dict

#####################################
###  generate fourier transforms  ###
def select_harmonic_transforms(harm_trf,dimensions,use_gpu):
    if (dimensions == 3) and use_gpu:
        trfs = harm_trf.transforms_by_indices['direct']
    elif (dimensions == 3) and (not use_gpu):
        trfs = harm_trf.transforms_by_indices['ml']
    elif  (dimensions == 2):
        trfs = harm_trf.transforms_by_indices['m']
    ht,iht = trfs['forward'],trfs['inverse']
    return ht,iht

def generate_ft(r_max,weights,harm_trf,dimensions,pos_orders=False,use_gpu=False,reciprocity_coefficient = np.pi,mode = 'trapz'):
    if not isinstance(pos_orders,bool):
        orders = pos_orders
    else:        
        orders=weights['posHarmOrders']
    weights_matrix = weights['weights']
    hankel,ihankel = generate_ht(weights_matrix,orders,r_max,reciprocity_coefficient=reciprocity_coefficient,dimensions=dimensions,use_gpu=use_gpu,mode=mode)
    ht,iht = select_harmonic_transforms(harm_trf,dimensions,use_gpu)
    def ft(data):
        #log.info('ft input shape = {}'.format(data.shape))
        harmonic_coefficients=ht(data)
        #log.info(f'coeff shape = {harmonic_coefficients.shape}')
        reciprocal_harmonic_coefficients=hankel(harmonic_coefficients)        
        reciprocal_function=iht(reciprocal_harmonic_coefficients)
        #reciprocal_function[0] = reciprocal_function[0].mean()
        #reciprocal_function[0] = 0
        #reciprocal_function[1] = reciprocal_function[1].mean()
        #reciprocal_function[2] = reciprocal_function[2].mean()
        #reciprocal_function[0]*= np.sqrt(2*np.pi)
        #reciprocal_function[0] = real_integrator(data)
        #reciprocal_function[0] = 0
        return reciprocal_function
    
    def ift(data):
        #log.info(f'data shape = {data.shape}')
        reciprocal_harmonic_coefficients=ht(data)
        #log.info(f'coeff shape = {len(reciprocal_harmonic_coefficients)}')
        harmonic_coefficients=ihankel(reciprocal_harmonic_coefficients)
        real_function=iht(harmonic_coefficients)
        #real_function[0] = real_function[0].mean()
        #real_function[0]*= np.sqrt(2*np.pi)
        #real_function[1] = real_function[1].mean()
        #real_function[2] = real_function[2].mean()
        #real_function[0] = reciprocal_integrator(data)
        #real_function[0] = 0
        #log.info('sum ift output = {}\n'.format(np.sum(real_function)))
        return real_function
    return ft,ift





import h5py as h5
import numpy as np
import logging
import traceback
from xframe.library.gridLibrary import NestedArray
from xframe.library.pythonLibrary import FTGridPair
from xframe.library.pythonLibrary import DictNamespace
from xframe.database.interfaces import HDF5Interface
log=logging.getLogger('root')

class HDF5_DB(HDF5Interface):
    h5=h5
    load_custom_types=False
    save_custom_types=False
    def __init__(self):
        save_custom_types={
            FTGridPair.__name__:HDF5_DB.save_FTGridPair,
            NestedArray.__name__:HDF5_DB.save_NestedArray,
            DictNamespace.__name__:HDF5_DB.save_DictNamespace
        }
        HDF5_DB.save_custom_types = save_custom_types

        load_custom_types={
            NestedArray.__name__:HDF5_DB.load_NestedArray
        }
        HDF5_DB.load_custom_types = load_custom_types
        
    @staticmethod
    def save(path,value_dict={},as_h5_object = False,**kwargs):     
        if isinstance(path,str):
            if as_h5_object:
                return h5.File(path, 'w', libver='latest')
            with h5.File(path, 'w') as h5file:
                HDF5_DB.recursively_save_dict_to_group(h5file,'/',value_dict)
        else:
            raise FileNotFoundError('Path Error, Abort loading "{}".'.format(path))
            
    @staticmethod
    def load(path,as_h5_object=False,**kwargs):
        try:
            if isinstance(path,str):
                if not as_h5_object:
                    with h5.File(path, 'r') as h5file:
                        data=HDF5_DB.recursively_load_dict_from_group(h5file, '/')
                else:
                    data = h5.File(path, 'r')
            else:
                raise FileNotFoundError('Path Error, Abort loading "{}".'.format(path))
        except Exception as e:
            raise e
        return data
    
    @staticmethod
    def recursively_save_dict_to_group(h5_file, path, dic):
        custom_save_routines = HDF5_DB.save_custom_types
        for key,item in dic.items():
            key=str(key)
            #log.info('key={}'.format(key))
            #log.info('path={}'.format(path))
            #log.info('h5-file={}'.format(h5_file.keys()))
            try:                
                if isinstance(item,(complex,float,int,bytes,bool,np.number,np.bool_)):
                    h5_file[path].create_dataset(key,data=item)
                elif isinstance(item,(str,np.character)):
                    h5_file[path].create_dataset(key,data=item.encode('utf-8'))
                    h5_file[path+'/'+key].attrs['type']='str'
                elif isinstance(item,np.ndarray):
                    HDF5_DB.save_numpy_array(h5_file,path,key,item)            
                elif isinstance(item,h5.VirtualLayout):
                    h5_file[path].create_virtual_dataset(key, item)
                elif isinstance(item,(list,tuple)):
                    #log.info(f'Saving list or tuple !!!! at {path+"/"+key}')
                    h5_file[path].create_group(key)
                    if isinstance(item,list):
                        h5_file[path+'/'+key].attrs['type']='list'
                    else:
                        h5_file[path+'/'+key].attrs['type']='tuple'                        
                    HDF5_DB.recursively_save_dict_to_group(h5_file, path + key + '/', {str(i):elem for i,elem in enumerate(item)})     
                elif isinstance(item, dict):
                    HDF5_DB._save_dict(h5_file,path,key,item)
                elif type(item).__name__ in custom_save_routines:
                    custom_save_routines[type(item).__name__](h5_file,path,key,item)
                else:
                    raise ValueError('Cannot save {} type for key {}'.format(type(item),key))
            except Exception as e:
                log.error("Failed to save key {} and item type {} with error: \n {}".format(key,type(item),e))
    
    @staticmethod
    def recursively_load_dict_from_group(h5_file, path):
        ans = {}
        custom_load_routines=HDF5_DB.load_custom_types
        for key, item in h5_file[path].items():
            item_type = item.attrs.get('type',False)
            if isinstance(item, h5._hl.dataset.Dataset):
                if item_type == 'str':
                    ans[key] = item[()].decode('utf-8')
                elif item_type in custom_load_routines:
                    ans[key] = custom_load_routines[item_type](item)
                else:                   
                    ans[key] = item[()]
            elif isinstance(item, h5._hl.group.Group):
                if item_type=='list':
                    ans[key] = HDF5_DB._load_list(h5_file,path,key)
                elif item_type=='tuple':
                    ans[key] = HDF5_DB._load_tuple(h5_file,path,key)
                else:
                    ans[key] = HDF5_DB.recursively_load_dict_from_group(h5_file, path + key + '/')
        return ans
    
    @staticmethod
    def save_numpy_array(h5_file,path,key,array,meta_dict={}):
        dtype=array.dtype
        if dtype==np.complex_:
            h5_file[path].create_dataset(key,data=array.astype('<c16'))
        elif dtype == np.bool_ :
                h5_file[path].create_dataset(key,data=array.astype(int))
        elif 'str' in dtype.name:
            h5_file[path].create_dataset(key,data=array.astype('S'))
        else:
            h5_file[path].create_dataset(key,data=array)
            for attrib in meta_dict:
                h5_file[path+key+'/'].attrs.create(attrib,meta_dict[attrib])


    @staticmethod
    def _save_dict(h5_file,path,key,item):        
        h5_file[path].create_group(key)
        #log.info('path key slash = {}'.format(path+key+'/') )
        HDF5_DB.recursively_save_dict_to_group(h5_file, path + key + '/', item)
    @staticmethod
    def _load_tuple(h5_file,path,key):
        data = HDF5_DB.recursively_load_dict_from_group(h5_file, path + key + '/')
        n_elements = len(data)
        return tuple(data[str(i)] for i in range(n_elements))
    @staticmethod
    def _load_list(h5_file,path,key):
        data = HDF5_DB.recursively_load_dict_from_group(h5_file, path + key + '/')
        n_elements = len(data)        
        return list(data[str(i)] for i in range(n_elements))

    ######## Custom Datattypes #######
    @staticmethod
    def save_NestedArray(h5_file,path,key,item):
        HDF5_DB.save_numpy_array(h5_file,path,key,item.array,meta_dict={'type':NestedArray.__name__,'n_ndim':item.n_ndim})
    @staticmethod
    def save_FTGridPair(h5_file,path,key,item):
        h5_file[path].create_group(key)
        new_item={'real_grid':item.realGrid,'reciprocal_grid':item.reciprocalGrid}
        HDF5_DB.recursively_save_dict_to_group(h5_file, path + key + '/', new_item)
    @staticmethod
    def save_DictNamespace(h5_file,path,key,item):
        HDF5_DB._save_dict(h5_file,path,key,item)
    @staticmethod
    def load_NestedArray(item):
        return NestedArray(item[()],item.attrs['n_ndim'])



import numpy as np
import logging

from xframe.library import units 
from xframe.library import mathLibrary as mLib
log=logging.getLogger('root')

#assumes scattering object is at the coordinate origin (0,0,0).
# beam is along z axis
def pixel_grid_to_scattering_grid(pixel_grid,wavelength,approximation='None',out_coord_sys='spherical'):
    if approximation == 'None':
        scattering_grid = get_spherical_scattering_grid(pixel_grid,wavelength)
    elif approximation == 'small_angle':
        scattering_grid = get_spherical_scattering_grid_small_angle(pixel_grid,wavelength)

    scattering_grid[...,0]*=units.standardLength
    if out_coord_sys == 'cartesian':
        scattering_grid = mLib.spherical_to_cartesian(scattering_grid)
    return scattering_grid
def get_pixel_areas_reziprocal(scattering_grid_cart,wavelength):
    g = scattering_grid_cart
    g[...,2] -= 2*np.pi/wavelength # shift ewald sphere center to origin.

    r = 2*np.pi/wavelength
    arc_angle_x = np.arccos(np.sum(g[:,:-1,:-1]*g[:,1:,:-1],axis = -1)/r**2)
    arc_angle_y = np.arccos(np.sum(g[:,:-1,:-1]*g[:,:-1,1:],axis = -1)/r**2)
    area = r**2*arc_angle_x*arc_angle_y
    return area
    
def get_spherical_scattering_grid_wrong(pixel_grid,incidentWavelength):
    detectorPixelGrid=pixel_grid

    radii=np.linalg.norm(detectorPixelGrid,axis=-1)
    zPositions=detectorPixelGrid[...,2]

    scatteringAngles = np.arctan2(zPositions,radii)

    reciprocalRadii = 2*np.sin(scatteringAngles/2)*1/incidentWavelength
    reciprocalTheta = (np.pi - scatteringAngles)/2
    reciprocalPhi = mLib.phiFrom_R_Theta_X_Y(radii,reciprocalTheta,detectorPixelGrid[...,0],detectorPixelGrid[...,1])                
    
    scattering_grid = np.stack((reciprocalRadii,reciprocalTheta,reciprocalPhi),axis=-1)
    return scattering_grid


def get_spherical_scattering_grid(pixel_grid,incidentWavelength):
    #pixel_grid[...,2]=100
    r = np.linalg.norm(pixel_grid,axis = -1)
    z = pixel_grid[...,2]
    scatteringAngles = np.zeros(z.shape)
    neg_z = z<0
    zr = z/r
    scatteringAngles[~neg_z] = np.arccos(zr[~neg_z])
    scatteringAngles[neg_z] = np.pi - np.arccos(-zr[neg_z])
    #print(scatteringAngles)
    reciprocalRadii = 4*np.pi*np.sin(scatteringAngles/2)/incidentWavelength
    #print(reciprocalRadii)
    reciprocalTheta = (np.pi-scatteringAngles)/2
    #log.info(reciprocalTheta)
    reciprocalPhi = np.arctan2(pixel_grid[...,1],pixel_grid[...,0])
    
    scattering_grid = np.stack((reciprocalRadii,reciprocalTheta,reciprocalPhi),axis=-1)
    return scattering_grid

def scattering_angle_to_reciprocal_radii(scattering_angle,xray_wavelength):
    ''' scattering_angle = 2*theta '''
    reciprocalRadii = 4*np.pi*np.sin(scattering_angle/2)/xray_wavelength
    return reciprocalRadii
def get_spherical_scattering_grid_small_angle(pixel_grid,incidentWavelength):
    r = np.linalg.norm(pixel_grid,axis = -1)
    r_azim = np.linalg.norm(pixel_grid[...,:2],axis = -1)
    z = pixel_grid[...,2]
    neg_z = z<0
    rr = r_azim/r
    scatteringAngles = np.zeros(z.shape)
    scatteringAngles[~neg_z] = rr[~neg_z]
    scatteringAngles[neg_z] = (np.pi-rr)[neg_z]
        
    reciprocalRadii=scatteringAngles*2*np.pi/incidentWavelength
    reciprocalTheta=(np.pi-scatteringAngles)/2
    reciprocalPhi = np.arctan2(pixel_grid[...,1],pixel_grid[...,0])
        
    scattering_grid=np.stack((reciprocalRadii,reciprocalTheta,reciprocalPhi),axis=-1)
    return scattering_grid


def ewald_sphere_theta_func(incidentWavelength):
    def theta(q):
        value=np.arccos(q*incidentWavelength/2)
        return value
    return theta


def ewald_sphere_theta_pi(incidentWavelength,qs):
    return np.arccos(qs*incidentWavelength/(4*np.pi)) 

def ewald_sphere_theta(incidentWavelength,qs):
    return np.arccos(qs*incidentWavelength/2)

def ewald_sphere_q_pi(wavelength,theta):
    return 4*np.pi*np.cos(theta)/wavelength
def ewald_sphere_wavelength_pi(max_q,theta):
    return 4*np.pi*np.cos(theta)/max_q

def energy_to_wavelength(energy):
    wavelength = (units.c*units.h)/energy
    return wavelength



def spherical_formfactor(q, radius = 1000):
    R = radius
    f = np.zeros_like(q)
    zero_mask = (q == 0.)
    V = (4/3)*np.pi*(R**3)
    f[zero_mask]= (4*np.pi*V**2)
    qq=q[~zero_mask]
    f[~zero_mask]= 36*np.pi*V**2*((np.sin(qq*R)-qq*R*np.cos(qq*R))/(qq*R)**3)**2    
    return f




import sys        
import logging
import numpy as np
import time

from PySide2 import QtCore      # core Qt functionality
from PySide2 import QtWidgets
from PySide2 import QtGui # extends QtCore with GUI functionality
from PySide2.QtOpenGL import QGLWidget as QOpenGLWidget    # provides QGLWidget, a special OpenGL QWidget
from PySide2.QtWidgets import QWidget,QStyleOptionSlider,QSizePolicy,QSlider,QStyle,QApplication,QCheckBox,QLabel,QHBoxLayout,QTextEdit,QPushButton
from PySide2.QtGui import QPainter,QPalette,QBrush,QPaintEvent,QMouseEvent,QKeySequence
from PySide2.QtCore import Qt,QSize,QRect
from PySide2 import QtOpenGL
from matplotlib import pyplot as plt
from matplotlib import cm
import OpenGL.GL as gl        # python wrapping of OpenGL
from OpenGL import GLU        # OpenGL Utility Library, extends OpenGL functionality
from OpenGL.arrays import vbo



log = logging.getLogger('root')
class RangeSlider(QSlider):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setValue(0.0)
        self.first_position = 0
        self.second_position = 100

        self.opt = QStyleOptionSlider()
        self.opt.minimum = 0.0
        self.opt.maximum = 100.0

        self.setTickPosition(QSlider.TicksBelow)
        self.setTickInterval(1)

        self.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed, QSizePolicy.Slider)
        )

    def setRangeLimit(self, minimum: int, maximum: int):
        self.opt.minimum = minimum
        self.opt.maximum = maximum

    def setRange(self, start: int, end: int):
        self.first_position = start
        self.second_position = end

    def getRange(self):
        return (self.first_position, self.second_position)

    def setTickPosition(self, position: QSlider.TickPosition):
        self.opt.tickPosition = position

    def setTickInterval(self, ti: int):
        self.opt.tickInterval = ti

    def paintEvent(self, event: QPaintEvent):

        painter = QPainter(self)

        # Draw rule
        self.opt.initFrom(self)
        self.opt.rect = self.rect()
        self.opt.sliderPosition = 0
        self.opt.subControls = QStyle.SC_SliderGroove | QStyle.SC_SliderTickmarks

        #   Draw GROOVE
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

        #  Draw INTERVAL

        color = self.palette().color(QPalette.Highlight)
        color.setAlpha(160)
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.NoPen)

        self.opt.sliderPosition = self.first_position
        x_left_handle = (
            self.style()
            .subControlRect(QStyle.CC_Slider, self.opt, QStyle.SC_SliderHandle)
            .right()
        )

        self.opt.sliderPosition = self.second_position
        x_right_handle = (
            self.style()
            .subControlRect(QStyle.CC_Slider, self.opt, QStyle.SC_SliderHandle)
            .left()
        )

        groove_rect = self.style().subControlRect(
            QStyle.CC_Slider, self.opt, QStyle.SC_SliderGroove
        )

        selection = QRect(
            x_left_handle,
            groove_rect.y(),
            x_right_handle - x_left_handle,
            groove_rect.height(),
        ).adjusted(-1, 1, 1, -1)

        painter.drawRect(selection)

        # Draw first handle

        self.opt.subControls = QStyle.SC_SliderHandle
        self.opt.sliderPosition = self.first_position
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

        # Draw second handle
        self.opt.sliderPosition = self.second_position
        self.style().drawComplexControl(QStyle.CC_Slider, self.opt, painter)

    def mousePressEvent(self, event: QMouseEvent):
        self.opt.sliderPosition = self.first_position
        pos = QtCore.QPoint()
        position = event.pos()
        pos.setX(int(position.x()))
        pos.setY(int(position.y()))
        self._first_sc = self.style().hitTestComplexControl(
            QStyle.CC_Slider, self.opt, pos, self
        )

        self.opt.sliderPosition = self.second_position
        self._second_sc = self.style().hitTestComplexControl(
            QStyle.CC_Slider, self.opt, pos, self
        )        

    def mouseMoveEvent(self, event: QMouseEvent):

        distance = self.opt.maximum - self.opt.minimum
        position = event.pos()
        pos = self.style().sliderValueFromPosition(
            0, distance, position.x(), self.rect().width()
        )

        self.trigger_value_change()
        if self._first_sc == QStyle.SC_SliderHandle:
            if pos <= self.second_position:
                self.first_position = pos
                self.update()
                return

        if self._second_sc == QStyle.SC_SliderHandle:
            if pos >= self.first_position:
                self.second_position = pos
                self.update()

    def trigger_value_change(self):
        v = self.value()
        if v == 0.0:
            self.setValue(1.0)
        else:
            self.setValue(0.0)
            
    def sizeHint(self):
        """ override """
        SliderLength = 84
        TickSpace = 5

        w = SliderLength
        h = self.style().pixelMetric(QStyle.PM_SliderThickness, self.opt, self)

        if (
            self.opt.tickPosition & QSlider.TicksAbove
            or self.opt.tickPosition & QSlider.TicksBelow
        ):
            h += TickSpace

        return (
            self.style()
            .sizeFromContents(QStyle.CT_Slider, self.opt, QSize(w, h), self)
#            .expandedTo(QApplication.globalStrut())
        )

    
class GLWidget(QOpenGLWidget):
    def __init__(self,vertices, indices, pixel_mask, init_data = False, parent=None):
        QOpenGLWidget.__init__(self, parent)
        
        if isinstance(init_data,bool) and (not isinstance(pixel_mask,bool)) :
            self._data = np.full(pixel_mask.shape,0.0)
            self._data[pixel_mask]=1.0    
        elif isinstance(init_data,np.ndarray):
            self._data = np.zeros(pixel_mask.shape)
            self._data[pixel_mask] = init_data.flatten()
        elif isinstance(vertices,bool) or isinstance(indices,bool) or isinstance(pixel_mask,bool):
            vertices,indices,init_data,pixel_mask = self.initGeometry()
            self._data = init_data
            
        self.mouse_is_dragged = False
        self.background_color = (0.2,0.2,0.2,0.0)
        self.low_clip_color = (0.4,0.4,0.4,0.0)
        self.high_clip_color = (0.0,1,0.0,0.0)
        self.view_angle = 45.0
        self._use_log_color_scale = True
        self._use_absolute_color_range = False
        self._absolute_color_limits = [0,100]
        self.current_color_limits = self._absolute_color_limits
        self._colormap= cm.get_cmap('plasma')
        self.parent = parent
        self.pixel_mask = pixel_mask
        #vertices,indices,colors = self.initGeometry()
        self._vertices = vertices
        self.create_view_port()
       
        self._colormap_range = (0.,1.)
        self._colors = self.data_to_color()
        self._indices = indices
        #print('shapes v {} i {} c {}'.format(self._vertices.shape,self._indices.shape, self._colors.shape))
        self.create_view_port()
        self.create_vbo()
        self.rendered_frames = 0
        self.fps_last_checked = time.time() 
        
        #gl.glClearColor(QtGui.QColor(0, 0, 255,255))    # initialize the screen to blue
        
 
    def create_view_port(self):
        verts = self.vertices
        center = np.mean(verts,axis = 0)
        #print('center = {}'.format(center))
        new_verts = verts - center
        max_dimension = np.max(np.linalg.norm(new_verts, axis = -1))
        self.max_dimension = max_dimension
        #print('max dimension = {}'.format(max_dimension))
        min_distance = (2*max_dimension)/np.tan(self.view_angle/360*2*np.pi)
        
        #print('min distance = {}'.format(min_distance))
        margin=-1
        max_distance = min_distance + (2+margin)*max_dimension
        #print('max distance = {}'.format(max_distance))
        object_z_shift = min_distance + (2+margin)/2*max_dimension
        object_shift = [0.0, 0.0, -object_z_shift]
        #print('object_distance distance = {}'.format(object_z_shift))
        self.viewport = [min_distance/100,max_distance*100,object_shift,-center]
        #print(self.viewport)                
        
        
                  
    def set_frame(self, vertices,data,indices):
        """Load 2D data as a Nx2 Numpy array.
        """
        self.vertices = vertices.astype(np.float32)
        self.data = data.astype(np.float32)
        self.indices = indices.astype(np.uint32)
        #vertices,indices,colors = self.initGeometry()
        #self.vertices = vertices
        #self.indices = indices
        
    def create_vc_vbo(self):
        vc = np.concatenate((self._vertices,self._colors),axis = 1).astype(np.float32)
        self.vc_VBO = vbo.VBO(vc)

    def create_index_vbo(self):
        self.index_VBO = vbo.VBO(self._indices,target='GL_ELEMENT_ARRAY_BUFFER')

    def data_to_color(self):
        if self._use_absolute_color_range:
            #print('use absolute color definition')
            colors = self.data_to_color_absolute()
        else:
            #print('use relative color definition')
            colors = self.data_to_color_relative()
        return colors
    def data_to_color_absolute(self):
        data = self._data
        length = self._absolute_color_limits[1] - self._absolute_color_limits[0]
        c_range = self._colormap_range
        view_min = self._absolute_color_limits[0] + c_range[0]*length
        view_max = self._absolute_color_limits[0] + c_range[1]*length
        view_length = view_max - view_min
        if view_length == 0:
            view_min = view_min*(1-1/2)
            view_max = view_max*(1+1/2)
            view_length = 1
        if self._use_log_color_scale:
            if view_max <=0:
                log.warning('All Values 0 or negative detected! They will be displayed as 0.')
                colors = np.full((len(data),4),0.0)
            else:
                if view_min<0:
                    length = self._absolute_color_limits[1]
                    view_min = c_range[0]*length
                    view_max = c_range[1]*length
                    view_length = view_max - view_min
                    log.warning('negative Values detected! They will be displayed as 0.')
                    
            view_data = 1+(data - view_min)/view_length*(np.exp(1)-1)
            low_data_mask = (view_data<=1)
            high_data_mask = (view_data>np.exp(1))
            view_data[low_data_mask] = 1.0
            view_data[high_data_mask] = np.exp(1)
            view_data = np.log(view_data)
            #print("view_data max = {} min = {}".format(view_data.max(),view_data.min()))
            colors = self._colormap(view_data)
            colors[low_data_mask,:] = self.low_clip_color
            colors[high_data_mask,:] = self.high_clip_color
        else:
            colors = self._colormap((data-view_min)/view_length)
        self.current_color_limits = [view_min,view_max]
        return colors.reshape(-1,4)

    def data_to_color_relative(self):
        data = self._data
        d_max = data.max()
        d_min = data.min()
        c_range = self._colormap_range
        length = d_max-d_min
        view_min = d_min + c_range[0]*length
        view_max = d_min + c_range[1]*length
        view_length = view_max - view_min
        if view_length == 0:
            view_min = view_min*(1-1/2)
            view_max = view_max*(1+1/2)
            view_length = 1
        if self._use_log_color_scale:
            if view_max <=0:
                log.warning('All Values 0 or negative detected! They will be displayed as 0.')
                colors = np.full((len(data),4),0.0)
            else:
                if view_min<0:
                    length = d_max
                    view_min = c_range[0]*length
                    view_max = c_range[1]*length
                    view_length = view_max - view_min
                    log.warning('negative Values detected! They will be displayed as 0.')

            view_data = 1+(data - view_min)/view_length*(np.exp(1)-1)
            low_data_mask = (view_data<=1)
            high_data_mask = (view_data>np.exp(1))
            view_data[low_data_mask] = 1.0
            view_data[high_data_mask] = np.exp(1)
            view_data = np.log(view_data)
            #print("view_data max = {} min = {}".format(view_data.max(),view_data.min()))
            colors = self._colormap(view_data)
            colors[low_data_mask,:] = self.low_clip_color
            colors[high_data_mask,:] = self.high_clip_color
        else:
            colors = self._colormap((data-view_min)/view_length)
        self.current_color_limits = [view_min,view_max]
        self._absolute_color_limits = [view_min,view_max]
        #print('colors shape ={}'.format(colors.shape))
        return colors.reshape(-1,4)
        
    def initializeGL(self):
        gl.glClearColor(*self.background_color)
        gl.glEnable(gl.GL_DEPTH_TEST)                  # enable depth testing
        gl.glShadeModel(gl.GL_FLAT)

        #self.initGeometry()
        #self.create_vbo()

        self.rotX = 180.0
        self.rotY = 0.0
        self.rotZ = 0.0
         
    def resizeGL(self, width, height):
        viewport = self.viewport
        gl.glViewport(0, 0, width, height)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        #gl.glOrtho(-500, 500, 500, -500, -2000, 2000)
        aspect = width / float(height)

        GLU.gluPerspective(self.view_angle, aspect, viewport[0], viewport[1])
        gl.glMatrixMode(gl.GL_MODELVIEW)

    def paintGL(self):
        self.print_fps()
        #start = time.time()
        viewport =self.viewport
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

        gl.glPushMatrix()    # push the current matrix to the current stack

        gl.glTranslate(*viewport[2])    # third, translate cube to specified depth
        #gl.glScale(20.0, 20.0, 20.0)       # second, scale cube
        gl.glRotate(self.rotX, 1.0, 0.0, 0.0)
        gl.glRotate(self.rotY, 0.0, 1.0, 0.0)
        gl.glRotate(self.rotZ, 0.0, 0.0, 1.0)
        gl.glTranslate(*viewport[3])   # first, translate cube center to origin

        self.vc_VBO.bind()
        #self.vbo.bind()
#        self.colorVBO.bind()
        self.index_VBO.bind()
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnableClientState(gl.GL_COLOR_ARRAY)
        
        gl.glColorPointer(4, gl.GL_FLOAT, 28, self.vc_VBO+12)
        gl.glVertexPointer(3, gl.GL_FLOAT, 28, self.vc_VBO)
        
        #gl.glDrawElements(gl.GL_TRIANGLES, len(self._indices), gl.GL_UNSIGNED_INT, None)
        gl.glDrawElements(gl.GL_QUADS, len(self._indices), gl.GL_UNSIGNED_INT, None)
        #gl.glDrawElements(gl.GL_POINTS, len(self._indices), gl.GL_UNSIGNED_INT, None)

        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDisableClientState(gl.GL_COLOR_ARRAY)

        gl.glPopMatrix()    # restore the previous modelview matrix
        #stop = time.time()
        #print('fps according to opengl = {}'.format(1/(stop-start)))
    def print_fps(self):
        rendered_frames = self.rendered_frames
        rendered_frames += 1
        current_time = time.time()
        elapsed_time = current_time-self.fps_last_checked
        if elapsed_time >=1:
            print('fps = {}'.format(rendered_frames))
            rendered_frames = 0
            self.fps_last_checked = current_time
        self.rendered_frames = rendered_frames
        
    def create_vbo(self):
        self.create_vc_vbo()
        self.create_index_vbo()

    def initGeometry(self):
        vertices = np.array(
                [[0.0, 0.0, 0.0],
                 [1.0, 0.0, 0.0],
                 [1.0, 1.0, 0.0],
                 [0.0, 1.0, 0.0],
                 [0.0, 0.0, 1.0],
                 [1.0, 0.0, 1.0],
                 [1.0, 1.0, 1.0],
                 [0.0, 1.0, 1.0]]).astype(np.float32)

        colors = np.array(
                [[0.0, 0.0, 0.0, 0.0],
                 [1.0, 0.0, 0.0, 0.0],
                 [1.0, 1.0, 0.0, 0.0],
                 [0.0, 1.0, 0.0, 0.0],
                 [0.0, 0.0, 1.0, 0.0],
                 [1.0, 0.0, 1.0, 0.0],
                 [1.0, 1.0, 1.0, 0.0],
                 [0.0, 1.0, 1.0, 0.0 ]]).astype(np.float32)
        data = np.arange(8,dtype = np.float32)
        mask = np.ones(8,dtype = bool)
        indices = np.array(
                [0, 1, 2, 3,
                 3, 2, 6, 7,
                 1, 0, 4, 5,
                 2, 1, 5, 6,
                 0, 3, 7, 4,
                 7, 6, 5, 4 ]).astype(np.int32)
        
        return vertices,indices,data,mask
        
    def setRotX(self, val):
        self.rotX = 360 * val + 180

    def setRotY(self, val):
        self.rotY = 360 * val

    def setRotZ(self, val):
        self.rotZ = 360 * val

    def addRotX(self, val):
        self.rotX += 360 * val
    def addRotY(self, val):
        self.rotY += 360 * val

    def addRotZ(self, val):
        self.rotZ += 360 * val

    
    def wheelEvent(self,event):
        min_distance = self.viewport[0]
        degree_change = event.angleDelta().y()/(8*360)*np.pi*2
        self.viewport[2][2] += degree_change*min_distance*10

    def keyPressEvent(self,event):
        step = self.max_dimension/10        
        if event.matches(Qt.Key_W):
            self.viewport[2][0] -= step
        elif event.matches(Qt.Key_S):
            self.viewport[2][0] -= step
        elif event.matches(Qt.Key_A):
            self.viewport[2][1] += step
        elif event.matches(Qt.Key_D):
            self.viewport[2][1] -= step
        
    def mouseMoveEvent(self, event):
        if self.mouse_is_dragged:
            x_start,y_start = self.mousepress_start_rotation_state[:2]
            old_pos = self.current_pos
            new_pos = event.pos()

            delta = [old_pos.x()-new_pos.x(), old_pos.y() - new_pos.y()]
            dampening = 10000
            x_rot_val = x_start + (-delta[1]/dampening)%1
            y_rot_val = (delta[0]/dampening)%1
        self.setRotX(x_rot_val)
        self.setRotY(y_rot_val)
            
    def mousePressEvent(self,event):
        self.current_pos=event.pos()
        self.mousepess_start_rotation_state = [self.rotX,self.rotY,self.rotZ]
        self.mouse_is_dragged=True
    def mouseReleaseEvent(self,event):
        self.mouse_is_dragged=False

    @property
    def vertices(self):
        return self._vertices
    
    @vertices.setter
    def vertices(self,vertices):
        self._vertices = vertices
        self.create_vc_vbo()        
        
    @property
    def indices(self):
        return self._indices
    @indices.setter
    def indices(self,indices):
        self._indices = indices
        self.create_vbo()
        
    @property
    def use_absolute_color_range(self):
        return self._use_absolute_color_range
    @use_absolute_color_range.setter
    def use_absolute_color_range(self,use_absolute_color_range):
        self._use_absolute_color_range = use_absolute_color_range
        #print('abs colors checkbox is changed')
        self.colors = self.data_to_color()

    @property
    def absolute_color_limits(self):
        return self._absolute_color_limits
    @absolute_color_limits.setter
    def absolute_color_limits(self,values):
        self._absolute_color_limits = values
        #print('abs colors range is changed')
        if self._use_absolute_color_range:
            #print('abs colors range is applied changed')
            self.colors = self.data_to_color()
        
    @property
    def data(self):
        return self.data
    @data.setter
    def data(self,data):
        self._data[self.pixel_mask] = data.flatten()
        self.colors = self.data_to_color()
        self.create_vc_vbo()

    @property
    def colors(self):
        return self.colors
    @colors.setter
    def colors(self,colors):
        self._colors = colors
        self.create_vc_vbo()

    @property
    def use_log_color_scale(self):
        return self._use_log_color_scale
    
    @use_log_color_scale.setter
    def use_log_color_scale(self,use_log_color_scale):
        self._use_log_color_scale = use_log_color_scale
        self.colors = self.data_to_color()

    @property
    def colormap(self):
        return self._colormap
    @colormap.setter
    def colormap(self,colormap_name):
        self._colormap = cm.get_cmap(colormap_name)
        self.colors = self.data_to_color()

    @property
    def colormap_range(self):
        return self._colormap_range
    @colormap_range.setter
    def colormap_range(self,c_range):
        color_range_changed = ~(np.isclose(c_range[0], self._colormap_range[0]) & np.isclose(c_range[1],self._colormap_range[1]))
        if color_range_changed:
            self._colormap_range = c_range
            self.colors = self.data_to_color()
            

class MainWindow(QtWidgets.QMainWindow):

    def __init__(self, pixel_vertices = False ,pixel_indices = False,sensitive_pixel_mask = 0):
        n_points=int(1e4)
        n_pics = int(1e3)
        if isinstance(pixel_vertices,bool):
            scale=0.2
            vertices = scale*np.random.randn(n_points,3).astype(np.float32)
            vertices[:,2]=0
            self.pixel_vertices = vertices
        else:
            self.pixel_vertices = pixel_vertices
            
        if isinstance(pixel_indices,bool):
            indices = np.arange(n_points,dtype=np.uint32)
            self.pixel_indices = indices
        else:
            self.pixel_indices = pixel_indices
        self.pixel_mask = sensitive_pixel_mask
        
        QtWidgets.QMainWindow.__init__(self)    # call the init for the parent class
        
        self.resize(300, 300)
        self.setWindowTitle('Hello OpenGL App')

        self._datasets = []
        self._meta_data=False
            
        self.glWidget = GLWidget(self.pixel_vertices, self.pixel_indices, self.pixel_mask)
        self.initGUI()
        
        timer = QtCore.QTimer(self)
        timer.setInterval(20)   # period, in milliseconds
        timer.timeout.connect(self.glWidget.update)
        timer.start()

    @property
    def meta_data(self):
        return meta_data
    @meta_data.setter
    def meta_data(self,_dict):
        self._meta_data = _dict
        
    @property
    def datasets(self):
        return self._datasets
    @datasets.setter
    def datasets(self,datasets):
        self._datasets = datasets
        self.pics_slider.setMaximum(max(0,len(datasets)-1))
        self.change_pic(0)
        
    def generate_data(self,n_points,n_pics):
        datasets = [np.random.rand(n_points).astype(np.float32) for i in range(n_pics)]
        return datasets
        
    def initGUI(self):
        central_widget = QtWidgets.QWidget()
        gui_layout = QtWidgets.QVBoxLayout()
        central_widget.setLayout(gui_layout)
        self.setCentralWidget(central_widget)
        
        gui_layout.addWidget(self.glWidget)

        sliderX = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        sliderX.valueChanged.connect(lambda val: self.glWidget.setRotX(float(val)/99))

        sliderY = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        sliderY.valueChanged.connect(lambda val: self.glWidget.setRotY(float(val)/99))

        sliderZ = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        sliderZ.valueChanged.connect(lambda val: self.glWidget.setRotZ(float(val)/99))

        slider_pics = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider_pics.valueChanged.connect(self.change_pic)
        slider_pics.setMinimum(0)
        slider_pics.setMaximum(max(0,len(self.datasets)-1))
        slider_pics.setSingleStep(1)
        self.pics_slider = slider_pics

        self.range_slider = RangeSlider()
        self.range_slider.valueChanged.connect(self.change_colormap_region)

        log_scale_check_box = QCheckBox('Log color scale')
        log_scale_check_box.stateChanged.connect(self.change_log_scale)
        log_scale_check_box.setChecked(True)

        title_label = QLabel()
        title_label.setAlignment(Qt.AlignHCenter | Qt.AlignBottom)
        title_label.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Maximum)
        self.title_label = title_label
        

        absolute_color_widget = QHBoxLayout()
        abs_col_check_box = QCheckBox('Absolute color range')
        abs_col_check_box.stateChanged.connect(self.change_use_abs_color_range)
        abs_col_min_text_edit = QTextEdit('')
        tmp = abs_col_min_text_edit
        textHeight = tmp.document().documentLayout().documentSize().height()
        tmp.setFixedHeight(textHeight + tmp.height() - tmp.viewport().height())
        self.abs_col_min_text_edit = abs_col_min_text_edit
        abs_col_max_text_edit = QTextEdit('')
        tmp = abs_col_max_text_edit
        abs_col_max_text_edit.setFixedHeight(textHeight + tmp.height() - tmp.viewport().height())
        self.abs_col_max_text_edit = abs_col_max_text_edit
        abs_color_apply_button = QPushButton('Apply')
        abs_color_apply_button.clicked.connect(self.change_abs_color_range)
        current_col_min_label = QLabel('Color Range Min:  ')
        self.current_col_min_label = current_col_min_label
        current_col_max_label = QLabel('Color Range Max:  ')
        self.current_col_max_label = current_col_max_label

        current_col_max_label.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        current_col_min_label.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        abs_col_max_text_edit.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        abs_col_min_text_edit.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        abs_col_check_box.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        abs_color_apply_button.setSizePolicy(QSizePolicy.Maximum,QSizePolicy.Maximum)
        
        absolute_color_widget.addWidget(abs_col_check_box)
        absolute_color_widget.addWidget(abs_col_min_text_edit)
        absolute_color_widget.addWidget(abs_col_max_text_edit)
        absolute_color_widget.addWidget(abs_color_apply_button)
        absolute_color_widget.addWidget(current_col_min_label)
        absolute_color_widget.addWidget(current_col_max_label)
                
        gui_layout.addWidget(title_label)
        gui_layout.addWidget(slider_pics)
        gui_layout.addLayout(absolute_color_widget)
        gui_layout.addWidget(self.range_slider)
        gui_layout.addWidget(log_scale_check_box)
        gui_layout.addWidget(sliderX)
        gui_layout.addWidget(sliderY)
        gui_layout.addWidget(sliderZ)

    def update_abs_color_range_text(self):
        limits = self.glWidget.absolute_color_limits
        self.abs_col_min_text_edit.setPlainText(str(limits[0]))
        self.abs_col_max_text_edit.setPlainText(str(limits[1]))                                           

    def update_current_color_range_label(self):
        limits = self.glWidget.current_color_limits
        self.current_col_min_label.setText('Color Range Min: {}'.format(str(limits[0])))
        self.current_col_max_label.setText('Color Range Max: {}'.format(str(limits[1])))
        
    def change_abs_color_range(self):
        _min = 0
        _max = 100
        try:
            _min = float(self.abs_col_min_text_edit.toPlainText())
            _max = float(self.abs_col_max_text_edit.toPlainText())
        except ValueError as e:
            log.error(e)
            log.warning('Using default absolute color min = {} ,max = {}'.format(_min,_max))
        if self.glWidget.use_log_color_scale:
            if _min < 0:
                log.warning(' min of absolute color range < 0. Is not applicable for logscale. Setting min = 0.')
                _min = 0
        self.glWidget.absolute_color_limits = [_min,_max]
        self.update_current_color_range_label()
                
                
    def change_use_abs_color_range(self,is_checked):        
        self.glWidget.use_absolute_color_range = bool(is_checked)
        self.update_abs_color_range_text()
        self.update_current_color_range_label()

    def change_colormap_region(self,*args):        
        c_range=self.range_slider.getRange()
        c_range = (float(c_range[0])/100,float(c_range[1])/100)
        self.glWidget.colormap_range=c_range
        self.update_current_color_range_label()
        
                
    def change_pic(self,val):
        datasets = self.datasets
        pic_id=int(val)
        self.change_title_label(pic_id)
        dataset = datasets[pic_id]
        self.glWidget.data = dataset
        self.change_colormap_region()
        if not self.glWidget.use_absolute_color_range:
            self.update_current_color_range_label()
            self.update_abs_color_range_text()
        
    def change_title_label(self, pic_id):
        if isinstance(self._meta_data,dict):
            text = ''
            tab = '    '
            for key,vals in self._meta_data.items():
                current_val = vals[pic_id]
                text += key + ': {}'.format(current_val) + tab
            text = text[:-len(tab)]
            self.title_label.setText(text)
            

    def keyPressEvent(self,event):
        gl_window = self.glWidget
        step = gl_window.max_dimension/50
        if event.text() == 'a':
            gl_window.viewport[2][0] -= step
        elif event.text() == 'd':
            gl_window.viewport[2][0] += step
        elif event.text() == 'w':
            gl_window.viewport[2][1] += step
        elif event.text() == 's':
            gl_window.viewport[2][1] -= step
        elif event.text() == 'h':
            gl_window.viewport[2][0] = 0.0
            gl_window.viewport[2][1] = 0.0


    def change_log_scale(self,is_checked):
        self.glWidget.use_log_color_scale = bool(is_checked)
        
class FramePresenter():
    def __init__(self,pixel_vertices=False,pixel_indices=False,sensitive_pixel_mask=False):
        if not QtWidgets.QApplication.instance():
            self.app = QtWidgets.QApplication(sys.argv)
        else:
            self.app = QtWidgets.QApplication.instance()
        self.win = MainWindow(pixel_vertices = pixel_vertices,pixel_indices = pixel_indices,sensitive_pixel_mask = sensitive_pixel_mask)
        
    def show(self,datasets, meta_data = False):
        self.win.meta_data = meta_data
        self.win.datasets = datasets
        self.win
        self.win.show() 
        self.app.exec_()
        #print('fi')



import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"

import numpy as np
import scipy
import logging
import traceback
from xframe import log

os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.library.gridLibrary import GridFactory,ReGrider,NestedArray
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft,zernike_radial_step_cutoff_relation, zernike_radial_cutoff_relation
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
log=log.setup_custom_logger('root','INFO')
import pdb_eda

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_model_0')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)

#f = db.load('pdb://2X53')
#db.save('test.vtr',[f['density']],grid = f['grid'][:], grid_type = 'cartesian')

a = pdb_eda.fromPDBid('3A12')
d = a.densityObj
h = d.header

crs_grid = GridFactory.construct_grid('uniform',[np.arange(n) for n in h.ncrs]).array
#crs_grid = np.transpose(crs_grid,axes = tuple(h.map2xyz)+(3,))
flattened_grid = crs_grid.reshape(-1,3)
flattened_grid += np.array(h.crsStart)[None,:]
flattened_grid = flattened_grid[:,h.map2xyz]/np.array(h.xyzInterval)[None,:]
xyz_shape = tuple(np.array(h.ncrs)[h.map2xyz])
xyz_grid = (np.dot(h.orthoMat,flattened_grid.T).T).reshape(tuple(h.ncrs)+(3,))
xyz_grid = np.moveaxis(xyz_grid,(0,1,2),h.map2xyz)
density = np.moveaxis(d.density, (0,1,2),h.map2xyz)

particle_size = 340
oversampling = 4
max_r = particle_size/2*oversampling


max_order = 99
n_radial_points = 128
opt=settings.analysis
ht_opt={'dimensions':3,'max_order':max_order,'indices':'lm','anti_aliazing_degree':2,'n_phi':False,'n_theta':False}
#ht_opt={'dimensions':opt.dimensions,**opt['grid'],'type':opt['fourier_transform']['type'],'pi_in_q':opt['fourier_transform']['pi_in_q'],'reciprocal_cut_off':maxQ}
# harmonic transforms and grid
cht=HarmonicTransform('complex',ht_opt)
weight_dict = generate_weightDict_zernike_spherical(max_order,n_radial_points)

max_q = zernike_radial_cutoff_relation(max_r,n_radial_points)

grid_opt={'dimensions':3,'type':'Zernike_dict','max_q':max_q,'n_radial_points':n_radial_points,**cht.grid_param,'pi_in_q':True}
grid_pair = get_grid(grid_opt)

#new_data = ReGrider.regrid(density,NestedArray(xyz_grid,1),'cartesian',NestedArray(grid_pair.real,1),'spherical',options={'fill_value': 0.0,'interpolation':'linear'})
db.save('test.vtr',[density],grid = xyz_grid, grid_type = 'cartesian')



import abc
import numpy as np
import logging
log = logging.getLogger('root')
from xframe.library.mathLibrary import SampleShapeFunctions,spherical_to_cartesian
import sys

class RegionOfInterest(abc.ABC):
    def __init__(self,parameters:dict,geometry:dict,modules=np.arange(16,dtype = int)):
        self.parameters = parameters
        self.geometry = geometry
        self.pixel_grid_spher = self.geometry['data_grid_spherical']
        #The following assumes spherical grid in form of r,theta,phi
        self.pixel_grid_polar = self.pixel_grid_spher[...,::2]        
        self.pixel_grid_cart = spherical_to_cartesian(self.pixel_grid_polar)
        self.n_total_modules = self.pixel_grid_cart.shape[0]
        self.data_modules = modules
        ## mask as if all of 16 modules would be used
        self.mask_complete = self.generate()
        self.mask_true_modules = self.calc_used_modules(self.mask_complete)
        self._used_modules = self.mask_true_modules
        ## mask data only using the data_modules
        self.used_module_ids = np.arange(len(self._used_modules))
        #log.info(modules)
        #log.info(self.mask_complete.shape)
        self.mask =  self.mask_complete[self._used_modules]

    @property
    def used_modules(self):
        return self._used_modules
    
    @used_modules.setter
    def used_modules(self,modules):
        #log.info('setter is called with {} '.format(modules))
        self._used_modules = modules
        self.used_module_ids = len(modules)
        #log.info('Mask shape before = {}'.format(self.mask.shape))
        self.mask = self.mask_complete[modules]
        #log.info('Mask shape after = {}'.format(self.mask.shape))
        
    @abc.abstractmethod
    def generate(self):
        ''' Template for generation routine. Has to return the mask and the used modules'''
        mask=None
        return mask

    @staticmethod
    def calc_used_modules(mask):
        used_modules = []
        for module,module_mask in enumerate(mask):
            if module_mask.any():
                used_modules.append(module)
        #return as tuple so that it throws an error if one tries to modify it
        return np.array(used_modules)
    

class Rectangle(RegionOfInterest):
    def generate(self):
        pixel_grid_cart = self.pixel_grid_cart
        
        x_len = self.parameters['x_len']
        y_len = self.parameters['y_len']
        center = self.parameters['center']        
        lengths = [x_len,y_len]
        rectangle_fct = SampleShapeFunctions.get_rectangle_function(lengths,center = center, coordSys = 'cartesian')
        
        mask = rectangle_fct(pixel_grid_cart).astype(bool)
        return mask

class Annulus(RegionOfInterest):
    def generate(self):
        pixel_grid_cart = self.pixel_grid_cart       

        inner_radius = self.parameters['inner_radius']
        outer_radius = self.parameters['outer_radius']
        center = self.parameters['center']
        anulus_fct = SampleShapeFunctions.get_anulus_function(inner_radius,outer_radius,center = center, coordSys = 'cartesian')
        
        mask = anulus_fct(pixel_grid_cart).astype(bool)        
        
        return mask

class Pixel(RegionOfInterest):
    def generate(self):
        pixel_grid_cart = self.pixel_grid_cart

        pixels = self.parameters['pixels']
        mask = np.zeros(pixel_grid_cart.shape[:-1],dtype = bool)
        index = tuple(zip(*pixels))
        mask[index] = True
        #log.info('pixels = {}'.format(pixels))
        #log.info('index array = {}'.format(index))
        #log.info('mask shape = {}'.format(mask.shape))                    
        
        return mask

class Asic(RegionOfInterest):
    def generate(self):
        pixel_grid_cart = self.pixel_grid_cart
        asics = self.parameters['asics']
        mask = np.zeros(pixel_grid_cart.shape[:-1],dtype = bool)
        asic_slices = self.geometry['asic_slices']
        for asic in asics:
            #log.info('asic ={}'.format(asic))
            slices = asic_slices[asic[1]][asic[2]]
            module = asic[0]
            mask[module,slices[0],slices[1]] = True
        return mask

class All(RegionOfInterest):
    def generate(self):
        mask = np.ones(self.n_total_modules,dtype = bool)
        return mask


class ROIManager():
    def __init__(self,geometry,rois_dict={},used_rois = [],data_modules=np.arange(16)):
        self.geometry=geometry
        self.rois = {'all':All({},geometry)}
        self._current_module = sys.modules[__name__]
        self.add_rois(rois_dict)
        self._used_modules = tuple()
        self._used_rois=[]
        self.used_rois=used_rois
        self._update_used_modules_from_used_rois()
        
    def add_rois(self,_dict):
        current_module = self._current_module
        for key,val in _dict.items():
            name = key
            _type = val['class']
            parameters = val['parameters']
            geometry = self.geometry
            try:
                #log.info(name)
                self.rois[name]= getattr(current_module,_type[0].upper()+_type[1:])(parameters,geometry)
            except AttributeError as e:
                log.warning('ROI type {}  of roi named {} not found.Continue.'.format(_type,name))
                
    def _update_used_modules_from_used_rois(self):
        modules = tuple()
        for name in self._used_rois:
            try:
                roi = self.rois[name]
                roi_modules = roi.mask_true_modules
                modules += tuple(roi_modules)                
            except KeyError as e:
                log.warning('ROI {} not found Known rois are {}.Skipping.'.format(name,self.rois.keys()))
        used_modules = np.unique(modules)
        self.used_modules = used_modules
        
    @property
    def used_modules(self):
        return self._used_modules
    @used_modules.setter
    def used_modules(self,modules):
        if len(modules)>0:
            self._used_modules = modules
            #1log.info(modules)
            for roi in self.rois.values():
                roi.used_modules = np.asarray(modules)
    @property
    def used_rois(self):
        return self._used_rois
    @used_rois.setter
    def used_rois(self,roi_names):
        new_used_rois = []
        for name in roi_names:
            if name in self.rois:
                new_used_rois.append(name)
        self._used_rois = new_used_rois
    
    def get_combined_roi_mask(self,roi_names):
        if len(roi_names)==0:
            mask = np.array(True)
        else:
            mask = False
            for name in roi_names:
                roi = self.rois[name]
                mask |= roi.mask
        return mask



import logging
log = logging.getLogger('root')
from xframe.presenters.interfaces import OpenCVInterface

CV_Plugin=OpenCVInterface

class Polar2D:
    @staticmethod
    def get_fig(*args,**kwargs):
        image=CV_Plugin.get_polar_image(*args,**kwargs)
        return image
    def get_fig_data(*args,**kwargs):
        image=CV_Plugin.get_polar_image_data(*args,**kwargs)
        return image

class Cart2D:
    @staticmethod
    def get_fig(*args,**kwargs):
        image=CV_Plugin.get_cart_image(*args,**kwargs)
        return image



import numpy as np
import logging

from xframe.library.gridLibrary import GridFactory
from xframe.library import pythonLibrary as pyLib
from xframe.library.interfaces import SphericalHarmonicTransformInterface
import shtns

log=logging.getLogger('root')

class sh(SphericalHarmonicTransformInterface):
    # want to use r,theta,phi type of data
    # l indexed coeff are f_{l,:} = coeff[l] 0<=l<=L_max of size: 2*l+1
    #   were coeff[l] cotains f_{l,m} starting with -l<=m<=l
    # m indexed coeff are f_{:,m} = coeff[m] -L_max<=m<=L_max of size: L_max-abs(m)
     
    
    def __init__(self,l_max,mode_flag='complex',output_order='l',anti_aliazing_degree = 2,n_phi = False,n_theta=False):
        #log.info('\n \n harmonic transform l_max = {} \n \n'.format(l_max)
        sh = shtns.sht(l_max)#,norm = shtns.sht_schmidt)        
        self._sh = sh
        self.l_max=l_max
        self.anti_aliazing_degree = anti_aliazing_degree        
        self.n_coeff = (l_max+1)**2
        #log.info(" sh created form n_phi= {},n_theta={},l_max ={}".format(n_phi,n_theta,l_max))
        thetas,phis,grid=self._generate_grid(n_phi=n_phi,n_theta=n_theta)
        #log.info(" sh created  grids n_phi= {},n_theta={}".format(len(phis),len(thetas)))
        self._theta=thetas
        self._phi=phis
        self._grid=grid
        self.mode=mode_flag
        self.m,self.l,self.cplx_m_indices,self.cplx_l_indices=self.generate_complex_lm_indices()
        self.cplx_m_indices_concat = np.concatenate(self.cplx_m_indices)
        self.cplx_l_split_indices = np.arange(1,l_max+1)**2
        self.cplx_m_split_indices = self.generate_cplx_m_split()
        self.real_split_indices=tuple((sh.l==l).nonzero()[0] for l in range(l_max+1))
        self.inv_real_split_indices=np.argsort(np.concatenate(self.real_split_indices))
        #log.info("alive!")
        if mode_flag=='real':
            self._forward_l=self.forward_transform_real_l
            self._inverse_l=self.inverse_transform_real_l
            self._forward_m=self.forward_transform_real_m
            self._inverse_m=self.forward_transform_real_m
            self._forward_direct = NotImplementedError()
            self._inverse_direct = NotImplementedError()
        elif mode_flag=='complex':
            self._forward_l=self.forward_transform_complex_l
            self._forward_m=self.forward_transform_complex_m
            self._inverse_l=self.inverse_transform_complex_l
            self._inverse_m=self.inverse_transform_complex_m
            self._forward_direct = self.generate_forward_transform_complex_direct()
            self._inverse_direct = self.generate_inverse_transform_complex_direct()
        else:
            raise AssertionError('harmonic transform mode "{}" is not known. Known modes are "real" and "complex".'.format(mode_flag))

    @property
    def phi(self):
        return self._phi
    @property
    def theta(self):
        return self._theta
    @property
    def grid(self):
        return self._grid

    @property
    def forward_l(self):
        return self._forward_l
    @property
    def forward_m(self):
        return self._forward_m
    @property
    def forward_d(self):
        return self._forward_direct
    @property
    def inverse_d(self):
        return self._inverse_direct
    @property
    def inverse_m(self):
        return self._inverse_m
    @property
    def inverse_l(self):
        return self._inverse_l

    def max_order_from_n_angular_steps(self,n_phi):
        '''
        max_order for highest power of 2 smaller than n_phi
        '''
        n_phi = 2**int(np.log2(n_phi))
        N = self.anti_aliazing_degree
        max_order = n_phi//(N+1)
        return max_order

    def n_angular_step_from_max_order(self,max_order):
        size_dict={}
        N = self.anti_aliazing_degree
        n_phi = 2**(int(np.log2((N+1)*max_order)) + 1)
        n_theta = n_phi//2
        size_dict['n_phi'] = n_phi
        size_dict['n_theta'] = n_theta
        return size_dict
    
    
    
    def generate_complex_lm_indices(self):
        l_max=self.l_max
        ls=np.arange(l_max+1,dtype=int)
        ms=np.concatenate((ls,-ls[:0:-1]))
        
        m_ordered_indices=[ls[np.abs(m):]*(ls[np.abs(m):]+1)+m for m in ms]

        l_ordered_indices=[slice(l**2,l**2+2*l+1) for l in  range(l_max+1)]                        

        return ms,ls,m_ordered_indices,l_ordered_indices


    def _generate_grid(self,n_phi=False,n_theta=False):
        sh=self._sh
        size_dict = self.n_angular_step_from_max_order(self.l_max)
        #log.info('n_theta = {}, n_phi = {}'.format(n_theta,n_phi))
        if (not isinstance(n_theta,int)) or isinstance(n_theta,bool):
            n_theta=size_dict['n_theta']
        if (not isinstance(n_phi,int)) or isinstance(n_phi,bool):
            n_phi = size_dict['n_phi']
            
        #with pyLib.stdout_redirected(): # external library has a frites directly to stdout this line suppresses it.
            #n_theta,n_phi = sh.set_grid(nlat = n_theta,nphi=n_phi,polar_opt=0,flags=shtns.sht_gauss)
        #log.info(f'nlat = {n_theta} nphi = {n_phi}')
        
        sh.set_grid(polar_opt=0,flags=shtns.sht_gauss) #extra call needed otherwise there are sometimes errors in shtns grid generation
        n_theta,n_phi = sh.set_grid(nlat = n_theta,nphi=n_phi,polar_opt=0,flags=shtns.sht_gauss)            
        phis=2*np.pi*np.arange(n_phi)/(n_phi*sh.mres)
        thetas=np.arccos(sh.cos_theta)
        grid=GridFactory.construct_grid('uniform',(thetas,phis))
        return thetas,phis,grid


    def _io_analysis_real_decorator(fun):        
        def new_function(self,data):
            sh=self._sh
            shape=data.shape[:-2]
            data=data.reshape(-1,sh.nlat,sh.nphi)
            #log.info('data shape ={} '.format(data.shape))
            val=fun(self,data)
            val=val.reshape(*shape,-1)
            #log.info('val_shape={}'.format(val.shape))
            val=tuple(val[:,indices] for indices in self.real_split_indices)
            return val
        return new_function
    
    def _analysis_decorator(fun):
        def new_function1(self,data):
            #log.info('forward shtns input shape= {}'.format(data.shape))
            #log.info('data pre shape = {}'.format(data.shape))
            sh=self._sh
            shape=data.shape[:-2]
            data=data.reshape(-1,*sh.spat_shape)
            val=fun(self,data)
            val=np.moveaxis(val.reshape(*shape,-1),-1,0)
            return val
        return new_function1
    
    @pyLib.optional_arg_decorator
    def _analysis_output_complex_decorator(fun,order='l'):
        if order=='l':            
            def new_function(self,data):
                val=fun(self,data)
                val=[ np.moveaxis(val[index],0,-1) for index in self.cplx_l_indices ]
                #log.info('l forward shtns output len= {}'.format(len(val)))
                return val
        elif order=='m':   
            def new_function(self,data):
                val=fun(self,data)
                val=[ np.moveaxis(val[index],0,-1) for index in self.cplx_m_indices ]
                #log.info('m forward shtns output len= {}'.format(len(val)))
                return val
        return new_function
    
    @pyLib.optional_arg_decorator
    def _synthesis_complex_decorator(fun,order='l'):
        if order=='l':
            def new_function(self,data):
                data=np.concatenate(data,axis=1)
                return fun(self,data)
        elif order=='m':
            def new_function(self,data):
                l_max=self.l_max
                lm_shape=(l_max**2+2*l_max+1,)
                r_shape=data[0].shape[:-1]                
                input_data=np.zeros(r_shape+lm_shape,dtype=complex)
                for m_id,index in enumerate(self.cplx_m_indices):
                    input_data[...,index]=data[m_id]
                return fun(self,input_data)
        return new_function

    def _io_synthesis_real_decorator(fun):
        def new_function(self,data):
            data=np.concatenate(data,axis=1)[...,self.inv_real_split_indices]
            return fun(self,data)
        return new_function


    @_io_analysis_real_decorator
    def forward_transform_real_l(self,data):
        return np.array(tuple(map(self._sh.analys,data)))
    @_io_synthesis_real_decorator
    def inverse_transform_real_l(self,data):
        return np.array(tuple(map(self._sh.synth,data)))


    def forward_transform_real_m(self,data):
        raise NotImplementedError()

    def inverse_transform_real_m(self,data):
        raise NotImplementedError()

    
    @_analysis_output_complex_decorator(order='l')
    @_analysis_decorator
    def forward_transform_complex_l(self,data):
        #data[0]=data[0].mean() # assumme r_0 = 0 
        return np.array(tuple(map(self._sh.analys_cplx,data)))
        

    @_analysis_output_complex_decorator(order='m')
    @_analysis_decorator
    def forward_transform_complex_m(self,data):
        #data[0]=data[0].mean() # assumme r_0 = 0 
        return np.array(tuple(map(self._sh.analys_cplx,data)))
    @_synthesis_complex_decorator(order='l')
    def inverse_transform_complex_l(self,data):
        #data[0,1:] = 0
        return np.array(tuple(map(self._sh.synth_cplx,data)))
    
    @_synthesis_complex_decorator(order='m')
    def inverse_transform_complex_m(self,data):
        #data[0,1:] = 0
        return np.array(tuple(map(self._sh.synth_cplx,data)))        


    def m_to_l_ordering(self,m_coeff):
        l_coeff = np.zeros_like(m_coeff)
        pos = 0
        for index in self.cplx_m_indices:
            n_parts = len(index)
            l_coeff[index] = m_coeff[pos:pos+n_parts]
            pos += n_parts
        return l_coeff

    def generate_forward_transform_complex_direct(self):
        analys_cplx = self._sh.analys_cplx
        def forward_transform_complex_direct(data):
            #data[0]=data[0].mean() # assumme r_0 = 0
            return np.array(tuple(analys_cplx(q_shell) for q_shell in data))
        return forward_transform_complex_direct

    def generate_inverse_transform_complex_direct(self):
        synth_cplx = self._sh.synth_cplx
        def inverse_transform_complex_direct(data):
            return np.array(tuple(synth_cplx(q_shell) for q_shell in data))
        return inverse_transform_complex_direct
    
    def _test_map(self,data):
        sh= self._sh
        return sh.synth_cplx(sh.analys_cplx(data))
    def test(self,data):
        return np.array(tuple(map(self._test_map,data+0.j)))        

    def generate_cplx_m_split(self):
        l_max = self.l_max
        lp1 = np.arange(l_max+2)
        index = (lp1*(lp1+1)/2).astype(int)
        split_ids = np.concatenate((index[-1] - index[-2::-1],index[-1]+index[1:-2]))
        return split_ids
